from .optimizer import Optimizer
from .parameter_range import (
    CategoricalRange,
    LogUniformFloatRange,
    LogUniformIntegerRange,
    ParameterRange,
    UniformFloatRange,
    UniformIntegerRange,
)

__all__ = [
    "ParameterRange",
    "UniformFloatRange",
    "LogUniformFloatRange",
    "UniformIntegerRange",
    "LogUniformIntegerRange",
    "CategoricalRange",
    "Optimizer",
]



import enum
import logging
import pickle
from typing import IO, TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import scipy.sparse as sps
from optuna import Study
from typing_extensions import Literal  # pragma: no cover, type: ignore

from .. import evaluation
from .._threading import get_n_threads
from ..definitions import (
    DenseMatrix,
    DenseScoreArray,
    InteractionMatrix,
    UserIndexArray,
)
from ..evaluation.evaluator import Evaluator
from ..optimization.parameter_range import (
    LogUniformFloatRange,
    ParameterRange,
    UniformIntegerRange,
)
from ._ials import IALSModelConfigBuilder, IALSSolverConfigBuilder
from ._ials import IALSTrainer as CoreTrainer
from ._ials import LossType, SolverType
from .base import (
    BaseRecommenderWithItemEmbedding,
    BaseRecommenderWithUserEmbedding,
    ParameterSuggestFunctionType,
)
from .base_earlystop import (
    BaseEarlyStoppingRecommenderConfig,
    BaseRecommenderWithEarlyStopping,
    TrainerBase,
)

if TYPE_CHECKING:
    from optuna import Study, Trial
    from optuna.storages import RDBStorage


def str_to_solver_type(t: str) -> SolverType:
    result: SolverType = getattr(SolverType, t.upper())
    assert result in {SolverType.CG, SolverType.CHOLESKY, SolverType.IALSPP}
    return result


def str_to_loss_type(t: str) -> LossType:
    result: LossType = getattr(LossType, t.upper())
    assert result in {LossType.ORIGINAL, LossType.IALSPP}
    return result


class IALSTrainer(TrainerBase):
    def __init__(
        self,
        X: InteractionMatrix,
        n_components: int,
        alpha0: float,
        reg: float,
        nu: float,
        init_std: float,
        solver_type: SolverType,
        max_cg_steps: int,
        ialspp_subspace_dimension: int,
        loss_type: LossType,
        random_seed: int,
        n_threads: int,
        prediction_time_max_cg_steps: int,
        prediction_time_ialspp_iteration: int,
    ):
        X_train_all_f32 = X.astype(np.float32)
        config = (
            IALSModelConfigBuilder()
            .set_K(n_components)
            .set_init_stdev(init_std)
            .set_alpha0(alpha0)
            .set_reg(reg)
            .set_nu(nu)
            .set_loss_type(loss_type)
            .set_random_seed(random_seed)
            .build()
        )
        solver_config = (
            IALSSolverConfigBuilder()
            .set_n_threads(n_threads)
            .set_solver_type(solver_type)
            .set_max_cg_steps(max_cg_steps)
            .set_ialspp_iteration(1)
            .set_ialspp_subspace_dimension(ialspp_subspace_dimension)
            .build()
        )
        self.core_trainer = CoreTrainer(config, X_train_all_f32)
        self.solver_config = solver_config
        self.prediction_time_solver_config = (
            IALSSolverConfigBuilder()
            .set_n_threads(n_threads)
            .set_solver_type(solver_type)
            .set_max_cg_steps(prediction_time_max_cg_steps)
            .set_ialspp_subspace_dimension(ialspp_subspace_dimension)
            .set_ialspp_iteration(prediction_time_ialspp_iteration)
            .build()
        )

    def load_state(self, ifs: IO) -> None:
        params = pickle.load(ifs)
        self.core_trainer.user = params["user"]
        self.core_trainer.item = params["item"]

    def compute_loss(self) -> float:
        return self.core_trainer.compute_loss(self.solver_config)

    def save_state(self, ofs: IO) -> None:
        pickle.dump(
            dict(user=self.core_trainer.user, item=self.core_trainer.item),
            ofs,
            protocol=pickle.HIGHEST_PROTOCOL,
        )

    def run_epoch(self) -> None:
        self.core_trainer.step(self.solver_config)

    def user_scores(self, begin: int, end: int) -> DenseScoreArray:
        return self.core_trainer.user_scores(begin, end, self.solver_config)

    def transform_user(self, X: InteractionMatrix) -> DenseMatrix:
        return self.core_trainer.transform_user(X, self.prediction_time_solver_config)

    def transform_item(self, X: InteractionMatrix) -> DenseMatrix:
        return self.core_trainer.transform_item(X, self.prediction_time_solver_config)


class IALSConfigScaling(enum.Enum):
    none = enum.auto()
    log = enum.auto()


class IALSConfig(BaseEarlyStoppingRecommenderConfig):
    n_components: int = 20
    alpha0: float = 1.0
    reg: float = 1e-3
    nu: float = 1.0
    confidence_scaling: str = "none"
    epsilon: float = 1.0
    init_std: float = 0.1
    solver_type: Literal["CG", "CHOLESKY"] = "CG"
    max_cg_steps: int = 3
    loss_type: Literal["IALSPP", "ORIGINAL"] = "IALSPP"
    nu_star: Optional[float] = None
    random_seed: int = 42
    n_threads: Optional[int] = None
    train_epochs: int = 16
    prediction_time_max_cg_steps: int = 5


def compute_reg_scale(X: sps.csr_matrix, alpha0: float, nu: float) -> float:
    X_csr: sps.csr_matrix = X.tocsr()
    X_csr.sort_indices()
    X_csc = X_csr.tocsc()
    X_csc.sort_indices()
    U, I = X.shape
    nnz_row: np.ndarray = X_csr.indptr[1:] - X_csr.indptr[:-1]
    nnz_col: np.ndarray = X_csc.indptr[1:] - X_csc.indptr[:-1]
    return float(((nnz_row + alpha0 * I) ** nu).sum()) + float(
        ((nnz_col + alpha0 * U) ** nu).sum()
    )


class IALSRecommender(
    BaseRecommenderWithEarlyStopping,
    BaseRecommenderWithUserEmbedding,
    BaseRecommenderWithItemEmbedding,
):
    r"""Implementation of implicit Alternating Least Squares (iALS) or Weighted Matrix Factorization (WMF).

    By default, it tries to minimize the following loss:

    .. math ::

        \frac{1}{2} \sum _{u, i \in S}  c_{ui} (\mathbf{u}_u \cdot \mathbf{v}_i - 1) ^ 2
        + \frac{\alpha_0}{2} \sum_{u, i} (\mathbf{u}_u \cdot \mathbf{v}_i) ^ 2 +
        \frac{\text{reg}}{2} \left( \sum_u (\alpha_0 I + N_u) ^ \nu || \mathbf{u}_u || ^2 +                            \sum_i (\alpha_0 U + N_i) ^ \nu || \mathbf{v}_i || ^2 \right)

    where :math:`S` denotes the set of all pairs wher :math:`X_{ui}` is non-zero.

    See the seminal paper:

        - `Collaborative filtering for implicit feedback datasets
          <http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.167.5120&rep=rep1&type=pdf>`_


    By default it uses a conjugate gradient descent version:

        - `Applications of the conjugate gradient method for implicit feedback collaborative filtering
          <https://dl.acm.org/doi/abs/10.1145/2043932.2043987>`_

    The loss above is slightly different from the original version. See the following paper for the loss used here

        - `Revisiting the Performance of iALS on Item Recommendation Benchmarks
          <https://arxiv.org/abs/2110.14037>`_

    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.
        n_components (int, optional):
            The dimension for latent factor. Defaults to 20.
        alpha0 (float, optional):
            The "unobserved" weight.
        reg (float, optional) :
            Regularization coefficient for both user & item factors. Defaults to 1e-3.
        nu (float, optional) :
            Controlles frequency regularization introduced in the paper,
            "Revisiting the Performance of iALS on Item Recommendation Benchmarks".
        confidence_scaling (str, optional) :
            Specifies how to scale confidence scaling :math:`c_{ui}`. Must be either "none" or "log".
            If "none", the non-zero (not-necessarily 1) :math:`X_{ui}` yields

            .. math ::
                c_{ui} = A + X_{ui}

            If "log",

            .. math ::
                c_{ui} = A + \log (1 + X_{ui} / \epsilon )

            The constant :math:`A` above will be 0 if ``loss_type`` is ``"IALSPP"``, :math:`\alpha_0` if ``loss_type`` is ``"ORIGINAL"``.

            Defaults to "none".
        epsilon (float, optional):
            The :math:`\epsilon` parameter for log-scaling described above.
            Will not have any effect if `confidence_scaling` is "none".
            Defaults to 1.0f.
        init_std (float, optional):
            Standard deviation for initialization normal distribution.
            The actual std for each user/item vector components are scaled by `1 / n_components ** .5`.
            Defaults to 0.1.
        solver_type ( "CHOLESKY" | "CG" | "IALSPP", optional):
            Which solver to use. Defaults to "CG".
        max_cg_steps (int, optional):
            Maximal number of conjute gradient descent steps during the training time.
            Defaults to 3.
            Used only when ``solver_type=="CG"``.
            By increasing this parameter, the result will be closer to
            Cholesky decomposition method (i.e., when ``solver_type == "CHOLESKY"``), but it wll take longer time.
        ialspp_subspace_dimension (int, optional):
            The subspace dimension of iALS++ (ignored if the ``solver_type`` is not "IALSPP").
            If this value is 1, specialized implementation described in `Fast Matrix Factorization for Online Recommendation with Implicit Feedback <https://arxiv.org/abs/1708.05024>`_ will be used instead.
            Defaults to 64.
        loss_type ( Literal["IALSPP", "ORIGINAL"], optional):
            Specifies the subtle difference between iALS++ vs Original Loss.
        nu_star (Optional[float], optional):
            If not `None`, used as the reference scale for nu described in the "Revisiting..." paper.
            Defaults to None.
        random_seed (int, optional):
            The random seed to initialize the parameters.
        n_threads (Optional[int], optional):
            Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if the variable is not set, it will be set to ``os.cpu_count()``. Defaults to None.
        train_epochs (int, optional):
            Maximal number of epochs. Defaults to 16.
        prediction_time_max_cg_steps (int, optional):
            Maximal number of conjute gradient descent steps during the prediction time,
            i.e., the case when a user unseen at the training time is given as a history matrix.
            Defaults to 5.

    Examples:

        >>> from irspack import IALSRecommender, rowwise_train_test_split, Evaluator
        >>> from irspack.utils.sample_data import mf_example_data
        >>> X = mf_example_data(100, 30, random_state=1)
        >>> X_train, X_test = rowwise_train_test_split(X, random_state=0)
        >>> rec = IALSRecommender(X_train)
        >>> rec.learn()
        >>> evaluator=Evaluator(X_test)
        >>> print(evaluator.get_scores(rec, [20]))
        OrderedDict([('hit@20', 1.0), ('recall@20', 0.9003412698412698), ('ndcg@20', 0.6175493479217139), ('map@20', 0.3848785870622406), ('precision@20', 0.3385), ('gini_index@20', 0.0814), ('entropy@20', 3.382497875272383), ('appeared_item@20', 30.0)])
    """

    config_class = IALSConfig
    default_tune_range: List[ParameterRange] = [
        UniformIntegerRange("n_components", 4, 300),
        LogUniformFloatRange("alpha0", 3e-3, 1),
        LogUniformFloatRange("reg", 1e-4, 1e-1),
    ]

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        n_components: int = 20,
        alpha0: float = 0.0,
        reg: float = 1e-3,
        nu: float = 1.0,
        confidence_scaling: str = "none",
        epsilon: float = 1.0,
        init_std: float = 0.1,
        solver_type: Literal["CG", "CHOLESKY", "IALSPP"] = "CG",
        max_cg_steps: int = 3,
        ialspp_subspace_dimension: int = 64,
        loss_type: Literal["IALSPP", "ORIGINAL"] = "IALSPP",
        nu_star: Optional[float] = None,
        random_seed: int = 42,
        n_threads: Optional[int] = None,
        train_epochs: int = 16,
        prediction_time_max_cg_steps: int = 5,
        prediction_time_ialspp_iteration: int = 7,
    ) -> None:

        super().__init__(
            X_train_all,
            train_epochs=train_epochs,
        )

        self.n_components = n_components
        self.alpha0 = alpha0
        self.reg = reg

        self.nu = nu
        self.confidence_scaling = IALSConfigScaling[confidence_scaling]
        self.epsilon = epsilon

        self.init_std = init_std
        self.solver_type = str_to_solver_type(solver_type)
        self.max_cg_steps = max_cg_steps
        self.ialspp_subspace_dimension = ialspp_subspace_dimension
        self.random_seed = random_seed
        self.n_threads = get_n_threads(n_threads)
        self.loss_type = str_to_loss_type(loss_type)
        self.nu_star = nu_star

        self.scaled_reg = self.reg
        if self.nu_star is not None:
            self.scaled_reg = (
                self.reg
                * compute_reg_scale(self.X_train_all, alpha0, self.nu_star)
                / compute_reg_scale(self.X_train_all, alpha0, nu)
            )
        self.prediction_time_max_cg_steps = prediction_time_max_cg_steps
        self.prediction_time_ialspp_iteration = prediction_time_ialspp_iteration

        self.trainer: Optional[IALSTrainer] = None

    @classmethod
    def _scale_X(
        cls, X: sps.csr_matrix, scheme: IALSConfigScaling, epsilon: float
    ) -> sps.csr_matrix:
        if scheme is IALSConfigScaling.none:
            return X
        else:
            X_ret: sps.csr_matrix = X.copy()
            X_ret.data = np.log(1 + X_ret.data / epsilon)
            return X_ret

    def _create_trainer(self) -> TrainerBase:
        return IALSTrainer(
            X=self._scale_X(self.X_train_all, self.confidence_scaling, self.epsilon),
            n_components=self.n_components,
            alpha0=self.alpha0,
            reg=self.scaled_reg,
            nu=self.nu,
            init_std=self.init_std,
            solver_type=self.solver_type,
            max_cg_steps=self.max_cg_steps,
            ialspp_subspace_dimension=self.ialspp_subspace_dimension,
            loss_type=self.loss_type,
            random_seed=self.random_seed,
            n_threads=self.n_threads,
            prediction_time_max_cg_steps=self.prediction_time_max_cg_steps,
            prediction_time_ialspp_iteration=self.prediction_time_ialspp_iteration,
        )

    @property
    def trainer_as_ials(self) -> IALSTrainer:
        if self.trainer is None:
            raise RuntimeError("tried to fetch trainer before the training.")
        return self.trainer

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        return self.trainer_as_ials.core_trainer.user[user_indices].dot(
            self.get_item_embedding().T
        )

    def get_score_block(self, begin: int, end: int) -> DenseScoreArray:
        return self.trainer_as_ials.user_scores(begin, end)

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        user_vector = self.compute_user_embedding(X)
        return self.get_score_from_user_embedding(user_vector)

    def get_user_embedding(self) -> DenseMatrix:
        return self.trainer_as_ials.core_trainer.user

    def get_score_from_user_embedding(
        self, user_embedding: DenseMatrix
    ) -> DenseScoreArray:
        return user_embedding.dot(self.get_item_embedding().T)

    def get_item_embedding(self) -> DenseMatrix:
        return self.trainer_as_ials.core_trainer.item

    def compute_user_embedding(self, X: InteractionMatrix) -> DenseMatrix:
        r"""Given an unknown users' interaction with known items,
        computes the latent factors of the users by least square (fixing item embeddings).

        parameters:
            X:
                The interaction history of the new users.
                ``X.shape[1]`` must be equal to ``self.n_items``.
        """
        return self.trainer_as_ials.transform_user(
            self._scale_X(
                sps.csr_matrix(X).astype(np.float32),
                self.confidence_scaling,
                self.epsilon,
            )
        )

    def compute_item_embedding(self, X: InteractionMatrix) -> DenseMatrix:
        r"""Given an unknown items' interaction with known user,
        computes the latent factors of the items by least square (fixing user embeddings).

        parameters:
            X:
                The interaction history of the new users.
                ``X.shape[0]`` must be equal to ``self.n_users``.
        """

        return self.trainer_as_ials.transform_item(
            self._scale_X(
                sps.csr_matrix(X).astype(np.float32),
                self.confidence_scaling,
                self.epsilon,
            )
        )

    def get_score_from_item_embedding(
        self, user_indices: UserIndexArray, item_embedding: DenseMatrix
    ) -> DenseScoreArray:
        return self.trainer_as_ials.core_trainer.user[user_indices].dot(
            item_embedding.T
        )

    @classmethod
    def tune_doubling_dimension(
        cls,
        data: InteractionMatrix,
        evaluator: Evaluator,
        initial_dimension: int,
        maximal_dimension: int,
        storage: Optional["RDBStorage"] = None,
        study_name_prefix: Optional[str] = None,
        n_trials_initial: int = 40,
        n_trials_following: int = 20,
        n_startup_trials_initial: int = 10,
        n_startup_trials_following: int = 5,
        max_epoch: int = 16,
        validate_epoch: int = 1,
        score_degradation_max: int = 3,
        neighborhood_scale: float = 3.0,
        suggest_function_initial: Optional[ParameterSuggestFunctionType] = None,
        random_seed: Optional[int] = None,
    ) -> Tuple[Dict[str, Any], pd.DataFrame]:
        r"""Perform tuning gradually doubling `n_components`.
        Typically, with the initial `n_components`, the search will be more exhaustive, and with larger `n_components`, less exploration will be done around previously found parameters.
        This strategy is described in `Revisiting the Performance of iALS on Item Recommendation Benchmarks <https://arxiv.org/abs/2110.14037>`_.

        Args:
            initial_dimension: The initial dimension.
            maximal_dimension: The maximal (inclusive) dimension to be tried.
            storage:
                The storage where multiple `optuna.Study` will be created corresponding to the various dimensions.
                If `None`, all `Study` will be created in-memory.
            study_name_prefix:
                The prefix for the names of `optuna.Study`. For dimension `d`, the full name of the `Study` will be `"{study_name_prefix}_{d}"`.
                If `None`, we will use a random string for this prefix.
            n_trials_initial:
                The number of trials for the initial dimension.
            n_trials_following:
                The number of trials for the following dimensions.
            n_startup_trials_initial:
                Passed on to `n_startup_trials` argument of `optuna.pruners.MedianPruner` in the initial `optuna.Study`.
                Defaults to `10`.
            n_startup_trials_following:
                Passed on to `n_startup_trials` argument of `optuna.pruners.MedianPruner` in the following `optuna.Study`.
                Defaults to `5`.
            neighborhood_scale:
                `alpha_0` and `reg` parameters will be searched within the log-uniform range
                [`previous_dimension_result / neighborhood_scale`, `previous_dimension_result * neighborhood_scale`].
                Defaults to `3.0`
            suggest_overwrite_initial:
                Overwrites the suggestion parameters in the initial `optuna.Study`.
                Defaults to `[]`.
            random_seed:
                The random seed to control ``optuna.samplers.TPESampler``. Defaults to `None`.

        Returns:
            A tuple that consists of
                1. A dict containing the best paramaters.
                   This dict can be passed to the recommender as ``**kwargs``.
                2. A ``pandas.DataFrame`` that contains the history of optimization for all dimensions.
        """
        from uuid import uuid1

        import optuna

        if study_name_prefix is None:
            study_name_prefix = str(uuid1())
        dimension = initial_dimension
        prev_params: Optional[Dict[str, Any]] = None
        results: List[Tuple[float, Dict[str, Any], pd.DataFrame]] = []
        while dimension <= maximal_dimension:
            _suggest: Optional[ParameterSuggestFunctionType]
            if random_seed is not None:
                random_seed += 1
            if prev_params is None:
                _suggest = suggest_function_initial
                n_trials = n_trials_initial
                n_startup_trials = n_startup_trials_initial
            else:
                n_trials = n_trials_following
                n_startup_trials = n_startup_trials_following

                _prev_params_copied = prev_params.copy()

                def _suggest(trial: "Trial") -> Dict[str, Any]:
                    result: Dict[str, Any] = {}
                    for key, value in _prev_params_copied.items():
                        if isinstance(value, float):
                            result[key] = trial.suggest_float(
                                key,
                                value / neighborhood_scale,
                                value * neighborhood_scale,
                                log=True,
                            )
                    return result

            study = optuna.create_study(
                storage=storage,
                sampler=optuna.samplers.TPESampler(seed=random_seed),
                pruner=optuna.pruners.MedianPruner(n_startup_trials=n_startup_trials),
                study_name=f"{study_name_prefix}_{dimension}",
            )
            bp, df_ = cls.tune_with_study(
                study,
                data,
                evaluator,
                n_trials,
                max_epoch=max_epoch,
                validate_epoch=validate_epoch,
                score_degradation_max=score_degradation_max,
                parameter_suggest_function=_suggest,
                fixed_params=dict(n_components=dimension),
            )
            df_["n_components"] = dimension
            results.append((float(df_["value"].min()), bp, df_))
            prev_params = study.best_params.copy()
            dimension *= 2
        final_result_df = pd.concat([x[2] for x in results])
        final_bp = sorted(results)[0][1]

        return final_bp, final_result_df

    @classmethod
    def tune_with_study(
        cls,
        study: "Study",
        data: Union[InteractionMatrix, None],
        evaluator: "evaluation.Evaluator",
        n_trials: int = 20,
        timeout: Optional[int] = None,
        data_suggest_function: Optional[Callable[["Trial"], InteractionMatrix]] = None,
        parameter_suggest_function: Optional[ParameterSuggestFunctionType] = None,
        fixed_params: Dict[str, Any] = dict(),
        max_epoch: int = 16,
        validate_epoch: int = 1,
        score_degradation_max: int = 3,
        logger: Optional[logging.Logger] = None,
    ) -> Tuple[Dict[str, Any], pd.DataFrame]:
        return super().tune_with_study(
            study,
            data,
            evaluator,
            n_trials,
            timeout,
            data_suggest_function,
            parameter_suggest_function,
            fixed_params,
            max_epoch,
            validate_epoch,
            score_degradation_max,
            logger,
        )

    @classmethod
    def tune(
        cls,
        data: Union[InteractionMatrix, None],
        evaluator: "evaluation.Evaluator",
        n_trials: int = 20,
        timeout: Optional[int] = None,
        data_suggest_function: Optional[Callable[["Trial"], InteractionMatrix]] = None,
        parameter_suggest_function: Optional[ParameterSuggestFunctionType] = None,
        fixed_params: Dict[str, Any] = dict(),
        random_seed: Optional[int] = None,
        prunning_n_startup_trials: int = 10,
        max_epoch: int = 16,
        validate_epoch: int = 1,
        score_degradation_max: int = 3,
        logger: Optional[logging.Logger] = None,
    ) -> Tuple[Dict[str, Any], pd.DataFrame]:
        return super().tune(
            data,
            evaluator,
            n_trials,
            timeout,
            data_suggest_function,
            parameter_suggest_function,
            fixed_params,
            random_seed,
            prunning_n_startup_trials,
            max_epoch,
            validate_epoch,
            score_degradation_max,
            logger,
        )



import re
from pathlib import Path
from typing import Tuple
from urllib.request import urlopen
from zipfile import ZipFile

import pandas as pd

from .downloader import BaseDownloader

_prefix = re.compile(r"\(\s*(\d+)\s*,\s*(\d+)\s*\)")


class NeuMFDownloader(BaseDownloader):
    TRAIN_URL: str
    NEGATIVE_URL: str
    _TRAIN_NAME = "train"
    _TEST_NAME = "test"

    def _save_to_zippath(self, path: Path) -> None:
        with ZipFile(path, "w") as save_zf:
            with save_zf.open(self._TRAIN_NAME, "w") as train_fs:
                b_train: bytes = urlopen(self.TRAIN_URL).read()
                train_fs.write(b_train)
            with save_zf.open(self._TEST_NAME, "w") as test_fs:
                b_test: bytes = urlopen(self.NEGATIVE_URL).read()
                test_fs.write(b_test)

    def read_train_test(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        with self.zf.open(self._TRAIN_NAME) as train_fs:
            train_df: pd.DataFrame = pd.read_csv(
                train_fs,
                header=None,
                sep="\t",
                names=["user_id", "item_id", "rating", "timestamp"],
            )
            train_df["timestamp"] = pd.to_datetime(train_df["timestamp"], unit="s")
        test_data = []
        with self.zf.open(self._TEST_NAME) as test_fs:
            for line_byte in test_fs:
                line = line_byte.decode()
                match_ = _prefix.search(line)
                assert match_ is not None
                uid_str, iid_str = match_.groups()
                uid = int(uid_str)
                iid = int(iid_str)
                test_data.append((uid, iid, True))
                for negative_iid in _prefix.sub("", line).strip().split("\t"):
                    test_data.append((uid, int(negative_iid), False))
        test_df = pd.DataFrame(
            test_data, columns=["user_id", "item_id", "positive"]
        ).drop_duplicates(["user_id", "item_id"], keep="first")
        return train_df, test_df


class NeuMFML1MDownloader(NeuMFDownloader):
    r"""Manages MovieLens 1M dataset split under 1-vs-100 negative evaluation protocol.

    Args:
        zippath:
            Where the zipped data is located. If `None`, assumes the path to be `~/.neumf-ml-1m.zip`.
            If the designated path does not exist, you will be prompted for the permission to download the data.
            Defaults to `None`.
        force_download:
            If `True`, the class will not prompt for the permission and start downloading immediately.
    """
    DEFAULT_PATH = Path("~/.neumf-ml-1m.zip").expanduser()

    TRAIN_URL = "https://raw.githubusercontent.com/tohtsky/neural_collaborative_filtering/master/Data/ml-1m.train.rating"
    NEGATIVE_URL = "https://raw.githubusercontent.com/tohtsky/neural_collaborative_filtering/master/Data/ml-1m.test.negative"


class NeuMFMPinterestDownloader(NeuMFDownloader):
    r"""Manages Pinterest dataset split under 1-vs-100 negative evaluation protocol.

    Args:
        zippath:
            Where the zipped data is located. If `None`, assumes the path to be `~/.neumf-pinterest.zip`.
            If the designated path does not exist, you will be prompted for the permission to download the data.
            Defaults to `None`.
        force_download:
            If `True`, the class will not prompt for the permission and start downloading immediately.
    """
    DEFAULT_PATH = Path("~/.neumf-pinterest.zip").expanduser()

    TRAIN_URL = "https://raw.githubusercontent.com/tohtsky/neural_collaborative_filtering/master/Data/pinterest-20.train.rating"
    NEGATIVE_URL = "https://raw.githubusercontent.com/tohtsky/neural_collaborative_filtering/master/Data/pinterest-20.test.negative"



import json
from typing import Any, Dict, List, Tuple, Type

import pandas as pd
from scipy import sparse as sps

from irspack import (
    AsymmetricCosineKNNRecommender,
    BaseRecommender,
    CosineKNNRecommender,
    DenseSLIMRecommender,
    EvaluatorWithColdUser,
    IALSRecommender,
    P3alphaRecommender,
    RP3betaRecommender,
    SLIMRecommender,
    TopPopRecommender,
    split_dataframe_partial_user_holdout,
)
from irspack.dataset import MovieLens20MDataManager

if __name__ == "__main__":

    BASE_CUTOFF = 100

    # We follow the preprocessing of Mult-VAE implementation (https://github.com/dawenl/vae_cf)
    data_manager = MovieLens20MDataManager()
    df_all = data_manager.read_interaction()
    df_all = df_all[df_all.rating >= 4]
    user_cnt = df_all.userId.value_counts()
    user_cnt = user_cnt[user_cnt >= 5]
    df_all = df_all[df_all.userId.isin(user_cnt.index)]

    data_all, _ = split_dataframe_partial_user_holdout(
        df_all,
        "userId",
        "movieId",
        n_test_user=10000,
        n_val_user=10000,
        heldout_ratio_val=0.2,
        heldout_ratio_test=0.2,
    )

    data_train = data_all["train"]
    data_val = data_all["val"]
    data_test = data_all["test"]

    X_train_val_all: sps.csr_matrix = sps.vstack(
        [data_train.X_all, data_val.X_all], format="csr"
    )
    valid_evaluator = EvaluatorWithColdUser(
        input_interaction=data_val.X_train,
        ground_truth=data_val.X_test,
        cutoff=BASE_CUTOFF,
    )
    test_evaluator = EvaluatorWithColdUser(
        input_interaction=data_test.X_train,
        ground_truth=data_test.X_test,
        cutoff=BASE_CUTOFF,
    )

    test_results = []
    validation_results = []

    test_configs: List[Tuple[Type[BaseRecommender], int, Dict[str, Any]]] = [
        (TopPopRecommender, 1, dict()),
        (CosineKNNRecommender, 40, dict()),
        (AsymmetricCosineKNNRecommender, 40, dict()),
        (P3alphaRecommender, 30, dict(alpha=1)),
        (RP3betaRecommender, 40, dict(alpha=1)),
        (DenseSLIMRecommender, 20, dict()),
        (SLIMRecommender, 40, dict()),  # Note: this is a heavy one.
    ]
    try:
        from irspack import MultVAERecommender

        test_configs.append(
            (
                MultVAERecommender,
                1,
                dict(
                    dim_z=200, enc_hidden_dims=600, kl_anneal_goal=0.2
                ),  # nothing to tune, use the parameters used in the paper.
            )
        )
    except ImportError:
        pass

    for recommender_class, n_trials, config in test_configs:
        recommender_name = recommender_class.__name__
        (best_param, validation_result_df) = recommender_class.tune(
            data_train.X_all,
            valid_evaluator,
            fixed_params=config,
            n_trials=n_trials,
            random_seed=0,
        )
        validation_result_df["recommender_name"] = recommender_name
        validation_results.append(validation_result_df)
        pd.concat(validation_results).to_csv(f"validation_scores.csv")
        test_recommender = recommender_class(X_train_val_all, **best_param).learn()
        test_scores = test_evaluator.get_scores(test_recommender, [20, 50, 100])

        test_results.append(
            dict(name=recommender_name, best_param=best_param, **test_scores)
        )
        with open("test_results.json", "w") as ofs:
            json.dump(test_results, ofs, indent=2)

    # Tuning following the strategy of
    # "Revisiting the Performance of iALS on Item Recommendation Benchmarks"
    # https://arxiv.org/abs/2110.14037
    (
        best_param_ials,
        validation_result_df_ials,
    ) = IALSRecommender.tune_doubling_dimension(
        data_train.X_all,
        valid_evaluator,
        initial_dimension=128,
        maximal_dimension=1024,
        random_seed=0,
        n_trials_initial=80,
        n_trials_following=40,
    )
    validation_result_df_ials["recommender_name"] = "IALSRecommender"
    validation_results.append(validation_result_df_ials)
    pd.concat(validation_results).to_csv(f"validation_scores.csv")
    test_recommender_ials = IALSRecommender(X_train_val_all, **best_param_ials)
    test_recommender_ials.learn()
    test_scores_ials = test_evaluator.get_scores(test_recommender_ials, [20, 50, 100])

    test_results.append(
        dict(name="IALSRecommender", best_param=best_param_ials, **test_scores_ials)
    )
    with open("test_results.json", "w") as ofs:
        json.dump(test_results, ofs, indent=2)



import math
from typing import Callable, Dict, Optional, Tuple

import numpy as np
import pytest
import scipy.sparse as sps

from irspack import Evaluator, rowwise_train_test_split
from irspack.recommenders import IALSRecommender


def ials_grad(
    X: sps.csr_matrix,
    u: np.ndarray,
    v: np.ndarray,
    reg: float,
    alpha0: float,
    epsilon: Optional[float] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    weight: Callable[[float], float]
    if epsilon is None:
        weight = lambda x: x
    else:
        eps_copy = epsilon
        weight = lambda x: math.log(1 + x / eps_copy)
    nu = u.shape[0]
    ni = v.shape[0]

    uv = u.dot(v.T)
    result_u = np.zeros_like(u)
    result_v = np.zeros_like(v)
    for uind in range(nu):
        for iind in range(ni):
            x = X[uind, iind]
            if x == 0:
                sc = alpha0 * uv[uind, iind]
            else:
                sc = (alpha0 + weight(x)) * (uv[uind, iind] - 1)

            result_u[uind, :] += v[iind] * sc
            result_v[iind, :] += u[uind] * sc
    result_u += reg * u
    result_v += reg * v
    return result_u, result_v


def test_ials_overfit_cholesky(
    test_interaction_data: Dict[str, sps.csr_matrix]
) -> None:
    X = test_interaction_data["X_small"]
    rec = IALSRecommender(
        X,
        n_components=4,
        alpha0=100,
        reg=1e-1,
        solver_type="CHOLESKY",
        loss_type="ORIGINAL",
        train_epochs=100,
        n_threads=1,
        nu=0,
    )
    rec.learn()
    assert rec.trainer is not None
    uvec = rec.compute_user_embedding(X)
    ivec = rec.compute_item_embedding(X)
    X = X.toarray()
    X[X.nonzero()] = 1.0
    reprod = uvec.dot(ivec.T)
    np.testing.assert_allclose(reprod, X, rtol=1e-2, atol=1e-2)


def test_ials_loss_original(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    X = test_interaction_data["X_small"]

    rec = IALSRecommender(
        X,
        n_components=2,
        alpha0=0.1,
        reg=1e-1,
        solver_type="CHOLESKY",
        loss_type="ORIGINAL",
        train_epochs=2,
        n_threads=1,
        nu=0,
    ).learn()
    assert rec.trainer is not None
    uvec = rec.get_user_embedding()
    ivec = rec.get_item_embedding()
    ui = uvec.dot(ivec.T)
    row, col = X.nonzero()

    # bruteforce computation of iALS loss
    loss_manual = (X.data + rec.alpha0).dot((ui[row, col] - 1) ** 2)
    ui[row, col] = 0.0
    loss_manual += rec.alpha0 * ui.ravel().dot(ui.ravel())
    loss_manual += rec.reg * ((uvec**2).sum() + (ivec**2).sum())
    loss_manual /= 2
    assert rec.trainer.compute_loss() == pytest.approx(loss_manual)


def test_ials_loss_ialspp(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    X = test_interaction_data["X_small"]

    rec = IALSRecommender(
        X,
        n_components=2,
        alpha0=0.1,
        reg=1e-1,
        solver_type="CHOLESKY",
        loss_type="IALSPP",
        train_epochs=2,
        n_threads=1,
        nu=0,
    ).learn()
    assert rec.trainer is not None
    uvec = rec.get_user_embedding()
    ivec = rec.get_item_embedding()
    ui = uvec.dot(ivec.T)
    row, col = X.nonzero()

    # bruteforce computation of iALS loss
    loss_manual = (X.data).dot((ui[row, col] - 1) ** 2)
    loss_manual += rec.alpha0 * ui.ravel().dot(ui.ravel())
    loss_manual += rec.reg * ((uvec**2).sum() + (ivec**2).sum())
    loss_manual /= 2
    assert rec.trainer.compute_loss() == pytest.approx(loss_manual)


def test_ials_overfit_cg(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    X = test_interaction_data["X_small"]
    rec = IALSRecommender(
        X,
        n_components=3,
        alpha0=100,
        loss_type="ORIGINAL",
        reg=1e-1,
        solver_type="CG",
        max_cg_steps=3,
        train_epochs=100,
        nu=0,
    )
    rec.learn()
    assert rec.trainer is not None
    uvec = rec.compute_user_embedding(X.tocsr().astype(np.float32))
    ivec = rec.compute_item_embedding(X.tocsr().astype(np.float32))
    X_dense = X.toarray()
    X_dense[X_dense.nonzero()] = 1.0
    reproduced_user_vector = uvec.dot(ivec.T)
    np.testing.assert_allclose(reproduced_user_vector, X_dense, rtol=1e-2, atol=1e-2)

    X_reproduced = rec.get_score_cold_user(X)
    np.testing.assert_allclose(X_reproduced, X_dense, rtol=1e-2, atol=1e-2)

    with pytest.raises(ValueError):
        _ = rec.compute_item_embedding(X.T)

    reproduced_item_vector = rec.compute_item_embedding(X)
    X_reproduced_item = rec.get_score_from_item_embedding(
        np.arange(X.shape[0]), reproduced_item_vector
    )
    np.testing.assert_allclose(X_reproduced_item, X_dense, rtol=1e-2, atol=1e-2)


@pytest.mark.parametrize(["subspace_dimension"], [(1,), (2,), (3,), (4,)])
def test_ials_overfit_ialspp(
    subspace_dimension: int, test_interaction_data: Dict[str, sps.csr_matrix]
) -> None:
    X = test_interaction_data["X_small"].copy()
    ALPHA0 = 100
    REG = 1.0
    rec = IALSRecommender(
        X,
        n_components=4,
        alpha0=ALPHA0,
        reg=REG,
        solver_type="IALSPP",
        loss_type="ORIGINAL",
        train_epochs=300,
        n_threads=1,
        ialspp_subspace_dimension=subspace_dimension,
        nu=0,
    )
    rec.learn()
    assert rec.trainer is not None
    uvec = rec.get_user_embedding()
    ivec = rec.get_item_embedding()
    X = X.toarray()
    X[X.nonzero()] = 1.0
    reprod = uvec.dot(ivec.T)
    np.testing.assert_allclose(reprod, X, rtol=1e-2, atol=1e-2)


@pytest.mark.xfail
def test_ials_cg_underfit(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    X = test_interaction_data["X_small"]
    rec = IALSRecommender(
        X,
        n_components=4,
        alpha0=1e3,
        reg=1e-3,
        solver_type="CG",
        max_cg_steps=1,
        train_epochs=10,
        nu=0,
    )
    with pytest.raises(RuntimeError):
        _ = rec.trainer_as_ials.core_trainer.user
    rec.learn()
    assert rec.trainer is not None
    uvec = rec.trainer.core_trainer.user
    ivec = rec.trainer.core_trainer.item
    X_dense = X.toarray()
    X_dense[X_dense.nonzero()] = 1.0
    reprod = uvec.dot(ivec.T)
    np.testing.assert_allclose(reprod, X_dense, rtol=1e-2, atol=1e-2)


def test_ials_overfit_nonzero_alpha(
    test_interaction_data: Dict[str, sps.csr_matrix]
) -> None:
    ALPHA = 4.5
    REG = 3
    X = test_interaction_data["X_small"]
    rec_chol = IALSRecommender(
        X,
        n_components=4,
        alpha0=1 / ALPHA,
        reg=REG,
        solver_type="CHOLESKY",
        train_epochs=5,
    )
    rec_chol.learn()
    assert rec_chol.trainer is not None
    uvec_chol = rec_chol.compute_user_embedding(X.tocsr().astype(np.float32))
    ivec_chol = rec_chol.compute_item_embedding(X.tocsr().astype(np.float32))

    rec_cg = IALSRecommender(
        X,
        n_components=4,
        alpha0=1 / ALPHA,
        reg=REG,
        solver_type="CG",
        max_cg_steps=5,
        train_epochs=5,
    )
    rec_cg.learn()
    assert rec_cg.trainer is not None
    uvec_cg = rec_cg.compute_user_embedding(X.tocsr())
    ivec_cg = rec_cg.compute_item_embedding(X.tocsr())

    np.testing.assert_allclose(uvec_chol, uvec_cg, atol=1e-3, rtol=1e-4)
    np.testing.assert_allclose(ivec_chol, ivec_cg, atol=1e-3, rtol=1e-4)


def test_ials_overfit_cholesky_logscale(
    test_interaction_data: Dict[str, sps.csr_matrix]
) -> None:

    ALPHA0 = 2.4
    REG = 1.1
    EPSILON = 3.0
    N_COMPONENTS = 5
    X = test_interaction_data["X_small"]
    rec_chol = IALSRecommender(
        X,
        n_components=N_COMPONENTS,
        alpha0=ALPHA0,
        reg=REG,
        nu=0,
        nu_star=0,
        solver_type="CHOLESKY",
        loss_type="ORIGINAL",
        epsilon=EPSILON,
        confidence_scaling="log",
        train_epochs=200,
        init_std=1e-1,
    )
    rec_chol.learn()
    assert rec_chol.trainer is not None
    uvec_chol = rec_chol.get_user_embedding()
    ivec_chol_cold = rec_chol.compute_item_embedding(X)

    grad_uvec_chol, grad_ivec_chol = ials_grad(
        X, uvec_chol, ivec_chol_cold, REG, ALPHA0, EPSILON
    )

    np.testing.assert_allclose(grad_ivec_chol, np.zeros_like(grad_ivec_chol), atol=1e-5)
    np.testing.assert_allclose(grad_uvec_chol, np.zeros_like(grad_uvec_chol), atol=1e-5)


def test_ials_tuning_with_n_startup_trials(
    test_interaction_data: Dict[str, sps.csr_matrix]
) -> None:
    X = test_interaction_data["X_small"]
    X_tr, X_te = rowwise_train_test_split(X, random_state=0)
    cutoff = min(10, X.shape[1])
    bp, history = IALSRecommender.tune(
        X_tr,
        Evaluator(X_te, cutoff=cutoff),
        random_seed=0,
        prunning_n_startup_trials=20,
        n_trials=20,
    )
    assert history[f"ndcg@{cutoff}"].isna().sum() == 0
    assert "train_epochs" in bp
    assert bp["train_epochs"] <= 16


def test_ials_tuning_with_too_early_n_startup(
    test_interaction_data: Dict[str, sps.csr_matrix]
) -> None:
    X = test_interaction_data["X_small"]
    X_tr, X_te = rowwise_train_test_split(X, random_state=0)
    cutoff = min(10, X.shape[1])
    bp, history = IALSRecommender.tune(
        X_tr,
        Evaluator(X_te, cutoff=cutoff),
        random_seed=0,
        prunning_n_startup_trials=1,
        n_trials=20,
        max_epoch=3,
    )
    assert (~history[f"ndcg@{cutoff}"].isna()).sum() > 0
    assert "train_epochs" in bp
    assert bp["train_epochs"] <= 3



from typing import List

import numpy as np
import pytest
import scipy.sparse as sps

from irspack.dataset.movielens import MovieLens100KDataManager
from irspack.definitions import DenseScoreArray, UserIndexArray
from irspack.evaluation import Evaluator
from irspack.optimization.parameter_range import (
    CategoricalRange,
    LogUniformFloatRange,
    LogUniformIntegerRange,
    ParameterRange,
    UniformFloatRange,
    UniformIntegerRange,
)
from irspack.recommenders import BaseRecommender
from irspack.split import rowwise_train_test_split

X_small = sps.csr_matrix(
    (np.random.RandomState(42).rand(100, 32) > 0.8).astype(np.float64)
)
ml_100k_df = MovieLens100KDataManager(force_download=True).read_interaction()
_, user_index = np.unique(ml_100k_df.userId, return_inverse=True)
_, movie_index = np.unique(ml_100k_df.movieId, return_inverse=True)
X_large = sps.csr_matrix(
    (np.ones(ml_100k_df.shape[0], dtype=np.float64), (user_index, movie_index)),
)


class MockRecommender(BaseRecommender, register_class=False):
    default_tune_range: List[ParameterRange] = [
        UniformFloatRange("p1", 0, 1),
        LogUniformFloatRange("reg", 0.99, 1.01),
        UniformIntegerRange("I1", 100, 102),
        LogUniformIntegerRange("I2", 500, 502),
        CategoricalRange("flag", ["foo", "bar"]),
    ]

    def __init__(
        self,
        X: sps.csr_matrix,
        X_test: sps.csr_matrix,
        p1: float = 1,
        I1: int = 1,
        I2: int = 1,
        reg: float = 1.0,
        flag: str = "hoge",
    ):
        super().__init__(X)
        self.p1 = p1  # only p1 matters
        self.I1 = I1
        self.I2 = I2
        self.reg = reg
        self.answer = X_test
        self.flag = flag
        self.rns = np.random.RandomState(0)

    def _learn(self) -> None:
        pass

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        score = self.answer[user_indices] * self.p1
        score = score + 10 * self.rns.rand(*score.shape) * (1 - self.p1)
        return score


@pytest.mark.parametrize("X", [X_small, X_large])
def test_optimizer_by_mock(X: sps.csr_matrix) -> None:
    X_train, X_val = rowwise_train_test_split(X, test_ratio=0.5, random_state=0)
    evaluator = Evaluator(X_val, 0)

    config, _ = MockRecommender.tune(
        X_train, evaluator, n_trials=40, random_seed=42, fixed_params=dict(X_test=X_val)
    )
    assert config["p1"] >= 0.9
    assert (config["reg"] >= 0.99) and (config["reg"] <= 1.01)
    assert (config["I1"] >= 100) and (config["I1"] <= 102)
    assert (config["I2"] >= 500) and (config["I2"] <= 502)
    assert config["flag"] in ["foo", "bar"]


def test_data_suggest_function() -> None:
    X_train_, X_val = rowwise_train_test_split(X_large, random_state=0)
    evaluator = Evaluator(X_val, offset=0, cutoff=100)
    from optuna import Trial

    from ..mock_recommender import MockRecommender as MockedScoreRecommender

    assert X_train_.shape == X_val.shape

    def data_suggest_function(trial: Trial) -> sps.csr_matrix:
        p = trial.suggest_float("p", 0, 1)
        if p < 0.5:
            return X_val
        else:
            return X_train_

    def parameter_suggest_function(trial: Trial) -> dict:
        return {"scores": np.random.randn(*X_train_.shape)}

    with pytest.raises(ValueError):
        _, __ = MockedScoreRecommender.tune(
            None,
            evaluator,
            n_trials=10,
            random_seed=42,
            data_suggest_function=None,
            parameter_suggest_function=parameter_suggest_function,
        )
    bp, history = MockedScoreRecommender.tune(
        None,
        evaluator,
        n_trials=10,
        random_seed=42,
        data_suggest_function=data_suggest_function,
        parameter_suggest_function=parameter_suggest_function,
    )
    for i, row in history.iterrows():
        hit_value = row[f"hit@100"]
        if row["p"] < 0.5:
            assert hit_value == 0.0
        else:
            assert hit_value > 0.0



import logging
import re
import time
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type

import pandas as pd

from ..default_logger import get_default_logger
from .parameter_range import is_valid_param_name

if TYPE_CHECKING:
    from optuna import Study, Trial

    from ..evaluation import Evaluator
    from ..recommenders.base import BaseRecommender, InteractionMatrix

SparseMatrixSuggestFunction = Callable[["Trial"], "InteractionMatrix"]
ParameterSuggestFunction = Callable[["Trial"], Dict[str, Any]]


def add_score_to_trial(trial: "Trial", score: Dict[str, float], cutoff: int) -> None:
    score_history: List[Tuple[int, Dict[str, float]]] = trial.study.user_attrs.get(
        "scores", []
    )
    score_history.append(
        (
            trial.number,
            {
                f"{score_name}@{cutoff}": score_value
                for score_name, score_value in score.items()
            },
        )
    )
    trial.study.set_user_attr("scores", score_history)


def study_to_dataframe(study: "Study") -> pd.DataFrame:
    result_df: pd.DataFrame = study.trials_dataframe().set_index("number")

    # remove prefix
    result_df.columns = [
        re.sub(r"^(user_attrs|params)_", "", colname) for colname in result_df.columns
    ]

    trial_and_scores: List[Tuple[float, Dict[str, float]]] = study.user_attrs.get(
        "scores", []
    )
    score_df = pd.DataFrame(
        [x[1] for x in trial_and_scores],
        index=[x[0] for x in trial_and_scores],
    )
    score_df.index.name = "number"
    result_df = result_df.join(score_df, how="left")
    return result_df


class Optimizer:
    r"""
    Args:
        data_suggest_function:
            The train data.
        val_evaluator:
            The validation evaluator that measures the performance of the recommenders.
        logger:
            The logger used during the optimization steps. Defaults to `None`.
            If `None`, the default logger of irspack will be used.
        suggest_overwrite:
            Customizes (e.g. enlarging the parameter region or adding new parameters to be tuned)
            the default parameter search space defined by ``default_tune_range``
            Defaults to list().
        fixed_params (Dict[str, Any], optional):
            Fixed parameters passed to recommenders during the optimization procedure.
            If such a parameter exists in ``default_tune_range``, it will not be tuned.
            Defaults to dict().
        max_epoch (int, optional):
            The maximal number of epochs for the training. Defaults to 512.
        validate_epoch (int, optional):
            The frequency of validation score measurement. Defaults to 5.
        score_degradation_max (int, optional):
            Maximal number of allowed score degradation. Defaults to 5. Defaults to 5."""

    def __init__(
        self,
        data_suggest_function: SparseMatrixSuggestFunction,
        parameter_suggest_function: ParameterSuggestFunction,
        fixed_params: Dict[str, Any],
        val_evaluator: "Evaluator",
        logger: Optional[logging.Logger] = None,
        max_epoch: int = 128,
        validate_epoch: int = 5,
        score_degradation_max: int = 5,
    ):

        if logger is None:

            logger = get_default_logger()

        self.logger = logger
        self._data_suggest_function = data_suggest_function
        self._parameter_suggest_function = parameter_suggest_function
        self.val_evaluator = val_evaluator

        self.current_trial: int = 0
        self.best_val = float("inf")
        self.fixed_params = fixed_params

        self.max_epoch = max_epoch
        self.validate_epoch = validate_epoch
        self.score_degradation_max = score_degradation_max

    def objective_function(
        self, recommender_class: Type["BaseRecommender"]
    ) -> Callable[["Trial"], float]:
        r"""Returns the objective function that can be passed to ``optuna.Study`` .

        Returns:
            A callable that receives ``otpuna.Trial`` and returns float (like ndcg score).
        """

        def objective_func(trial: "Trial") -> float:
            start = time.time()
            data = self._data_suggest_function(trial)
            params = self._parameter_suggest_function(trial)
            params.update(self.fixed_params)
            self.logger.info("Trial %s:", trial.number)
            self.logger.info("parameter = %s", params)

            recommender = recommender_class(data, **params)
            recommender.learn_with_optimizer(
                self.val_evaluator,
                trial,
                max_epoch=self.max_epoch,
                validate_epoch=self.validate_epoch,
                score_degradation_max=self.score_degradation_max,
            )

            score = self.val_evaluator.get_score(recommender)
            end = time.time()

            time_spent = end - start
            self.logger.info(
                "Config %d obtained the following scores: %s within %f seconds.",
                trial.number,
                score,
                time_spent,
            )
            val_score = score[self.val_evaluator.target_metric.name]

            # max_epoch will be stored in learnt_config
            for param_name, param_val in recommender.learnt_config.items():
                trial.set_user_attr(param_name, param_val)

            add_score_to_trial(trial, score, self.val_evaluator.cutoff)

            return -val_score

        return objective_func

    def optimize_with_study(
        self,
        study: "Study",
        recommender_class: Type["BaseRecommender"],
        n_trials: int = 20,
        timeout: Optional[int] = None,
    ) -> Tuple[Dict[str, Any], pd.DataFrame]:
        """Perform the optimization step using the user-created ``optuna.Study`` object.
        Creating and managing the study object will be convenient e.g. when you

            1. want to `store/resume the study using RDB backend <https://optuna.readthedocs.io/en/stable/tutorial/003_rdb.html>`_.
            2. want perform a `distributed optimization <https://optuna.readthedocs.io/en/stable/tutorial/004_distributed.html>`_.

        Args:
            study:
                The study object.
            n_trials:
                The number of expected trials (include pruned trial.). Defaults to 20.
        Returns:
            A tuple that consists of

                1. A dict containing the best paramaters.
                   This dict can be passed to the recommender as ``**kwargs``.
                2. A ``pandas.DataFrame`` that contains the history of optimization.

        """

        objective_func = self.objective_function(recommender_class)

        self.logger.info(
            """Start parameter search for %s""",
            recommender_class.__name__,
        )

        study.optimize(objective_func, n_trials=n_trials, timeout=timeout)
        best_params = dict(
            **study.best_trial.params,
            **{
                key: val
                for key, val in study.best_trial.user_attrs.items()
                if is_valid_param_name(key)
            },
        )
        best_params.update(self.fixed_params)
        trials_df = study_to_dataframe(study)
        return (best_params, trials_df)



import sys
from datetime import datetime
from io import BytesIO
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pytest
from pytest_mock import MockFixture

from irspack.dataset import neu_mf

ZIPFILE_NAME = Path("~").expanduser() / "ml.test.zip"


class DummyNeuMFDownloader(neu_mf.NeuMFDownloader):
    TRAIN_URL: str = ""
    NEGATIVE_URL: str = ""


def test_newumf_dataset(mocker: MockFixture) -> None:
    if sys.platform == "win32":
        pytest.skip("Skip on Windows.")

    rng = np.random.default_rng(0)
    train_item_ids = {1: {1, 2, 3, 100}, 3: {2, 4, 101}, 4: {1, 4, 5, 102}}
    all_item_ids = sorted(list({z for y in train_item_ids.values() for z in y}))
    validation_true_positive = {1: 5, 3: 1, 4: 2}
    train_rows = []

    YEAR = 2021
    start_timestamp = datetime(YEAR, 1, 1).timestamp()

    for user_id, interacted_item_ids in train_item_ids.items():
        for item_id in interacted_item_ids:
            rating = (rng.integers(1, 5),)
            timestamp = int(start_timestamp) + 86400 * rng.integers(0, 364)
            train_rows.append(f"{user_id}\t{item_id}\t{rating}\t{timestamp}")
    train_bytes = "\n".join(train_rows).encode()
    test_bytes_list = []
    for user_id, interacted_item_ids in train_item_ids.items():
        validation_gt = validation_true_positive[user_id]
        untouched_item_set = (
            set(all_item_ids)
            .difference(interacted_item_ids)
            .difference({validation_gt})
        )
        samples = rng.choice(
            list(untouched_item_set),
            size=min(len(untouched_item_set), 3),
            replace=False,
        )
        test_bytes_list.append(
            f"({user_id},{validation_gt})\t" + "\t".join([f"{iid}" for iid in samples])
        )
    test_bytes = "\n".join(test_bytes_list).encode()

    mocker.patch(
        "irspack.dataset.neu_mf.urlopen",
        side_effect=[BytesIO(train_bytes), BytesIO(test_bytes)],
        autospec=True,
    )
    with TemporaryDirectory() as temp_dir:
        save_path = Path(temp_dir) / "hoge.zip"
        dm = DummyNeuMFDownloader(save_path, force_download=True)
        train, test = dm.read_train_test()
    assert np.all(train["timestamp"].dt.year.values == YEAR)
    for row in train.itertuples():
        user_id = int(row.user_id)
        item_id = int(row.item_id)
        assert item_id in train_item_ids[user_id]

    for row in test[test["positive"]].itertuples():
        user_id = int(row.user_id)
        item_id = int(row.item_id)
        assert item_id == validation_true_positive[user_id]

    for row in test[~test["positive"]].itertuples():
        user_id = int(row.user_id)
        item_id = int(row.item_id)
        assert item_id != validation_true_positive[user_id]
        assert item_id not in train_item_ids[user_id]



from collections import defaultdict
from typing import Any, List

import numpy as np
import pytest
from scipy import sparse as sps

from irspack.evaluation import Evaluator
from irspack.recommenders.base import BaseRecommender


class MockRecommender(BaseRecommender, register_class=False):
    def __init__(self, X_all: sps.csr_matrix, scores: np.ndarray) -> None:
        super().__init__(X_all)
        self.scores = scores

    def get_score(self, user_indices: np.ndarray) -> np.ndarray:
        return self.scores[user_indices]

    def _learn(self) -> None:
        pass


@pytest.mark.parametrize("U, I, R", [(10, 30, 10), (100, 30, 5), (30, 100, 2)])
def test_restriction_global(U: int, I: int, R: int) -> None:
    try:
        from sklearn.metrics import ndcg_score
    except:
        pytest.skip()
    rns = np.random.RandomState(42)
    recommendable = rns.choice(np.arange(I), replace=False, size=R)
    scores = rns.randn(U, I)
    X_gt = (rns.rand(U, I) >= 0.3).astype(np.float64)
    evaluator_ = Evaluator(
        sps.csr_matrix(X_gt),
        offset=0,
        cutoff=I,
        n_threads=1,
        recommendable_items=recommendable,
    )
    # empty mask
    mock_rec = MockRecommender(sps.csr_matrix(X_gt.shape), scores)
    my_score = evaluator_.get_score(mock_rec)
    sklearn_metrics = defaultdict(list)
    for i in range(scores.shape[0]):
        if X_gt[i, recommendable].sum() == 0:
            continue
        ndcg = ndcg_score(
            X_gt[i, recommendable][None, :], scores[i, recommendable][None, :]
        )
        sklearn_metrics["ndcg"].append(ndcg)

    assert my_score["ndcg"] == pytest.approx(np.mean(sklearn_metrics["ndcg"]), abs=1e-8)


@pytest.mark.parametrize(
    "U, I, is_sparse", [(10, 30, False), (100, 30, True), (30, 100, False)]
)
def test_restriction_local(U: int, I: int, is_sparse: bool) -> None:
    try:
        from sklearn.metrics import ndcg_score
    except:
        pytest.skip()
    rns = np.random.RandomState(42)
    recommendables: List[List[int]] = []

    def _to_int_list(x: Any) -> List[int]:
        return [int(_) for _ in x]

    for _ in range(U):
        recommendables.append(
            _to_int_list(
                rns.choice(np.arange(I), replace=False, size=rns.randint(2, I))
            )
        )
    if is_sparse:
        indices = [j for r in recommendables for j in sorted(r)]
        indptr = np.cumsum([0] + [len(r) for r in recommendables])
        recommendables_arg = sps.csr_matrix(
            (np.ones(len(indices), dtype=float), indices, indptr), shape=(U, I)
        )
    else:
        recommendables_arg = recommendables
    scores = rns.randn(U, I)
    X_gt = (rns.rand(U, I) >= 0.3).astype(np.float64)
    eval = Evaluator(
        sps.csr_matrix(X_gt),
        offset=0,
        cutoff=I,
        n_threads=1,
        per_user_recommendable_items=recommendables_arg,
    )
    # empty mask
    mock_rec = MockRecommender(sps.csr_matrix(X_gt.shape), scores)
    my_score = eval.get_score(mock_rec)
    sklearn_metrics = defaultdict(list)
    for i in range(scores.shape[0]):
        if X_gt[i, recommendables[i]].sum() == 0:
            continue
        ndcg = ndcg_score(
            X_gt[i, recommendables[i]][None, :], scores[i, recommendables[i]][None, :]
        )
        sklearn_metrics["ndcg"].append(ndcg)

    assert my_score["ndcg"] == pytest.approx(np.mean(sklearn_metrics["ndcg"]), abs=1e-8)


@pytest.mark.parametrize("U, I", [(10, 30)])
def test_irregular(U: int, I: int) -> None:
    rns = np.random.RandomState(42)
    recommendables: List[np.ndarray] = []
    X_gt = (rns.rand(U, I) >= 0.3).astype(np.float64)
    _ = Evaluator(
        sps.csr_matrix(X_gt),
        offset=0,
        cutoff=I,
        n_threads=1,
        per_user_recommendable_items=[[0] for _ in range(X_gt.shape[0])],
    )
    _ = Evaluator(
        sps.csr_matrix(X_gt),
        offset=0,
        cutoff=I,
        n_threads=1,
        per_user_recommendable_items=sps.csr_matrix(X_gt),
    )

    with pytest.raises(ValueError):
        _ = Evaluator(
            sps.csr_matrix(X_gt),
            offset=0,
            cutoff=I,
            n_threads=1,
            per_user_recommendable_items=[[0], [0]],
        )
    with pytest.raises(ValueError):
        _ = Evaluator(
            sps.csr_matrix(X_gt),
            offset=0,
            cutoff=I,
            n_threads=1,
            per_user_recommendable_items=[[0, 0]],
        )
    with pytest.raises(ValueError):
        _ = Evaluator(
            sps.csr_matrix(X_gt),
            offset=0,
            cutoff=I,
            n_threads=1,
            per_user_recommendable_items=[
                [
                    I,
                ]
            ],
        )
    with pytest.raises(ValueError):
        _ = Evaluator(
            sps.csr_matrix(X_gt),
            offset=0,
            cutoff=I,
            n_threads=1,
            per_user_recommendable_items=X_gt[:-1],
        )



import logging
from abc import ABCMeta, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    no_type_check,
)

import numpy as np
import pandas as pd
from pydantic import BaseModel
from scipy import sparse as sps

if TYPE_CHECKING:
    from optuna import Study, Trial

    from .. import evaluation


ParameterSuggestFunctionType = Callable[["Trial"], Dict[str, Any]]

from ..definitions import (
    DenseMatrix,
    DenseScoreArray,
    InteractionMatrix,
    UserIndexArray,
)
from ..optimization.parameter_range import ParameterRange

R = TypeVar("R", bound="BaseRecommender")


def _sparse_to_array(U: Any) -> np.ndarray:
    if sps.issparse(U):
        return U.toarray()
    else:
        return U


class CallBeforeFitError(Exception):
    pass


class RecommenderConfig(BaseModel):
    class Config:
        extra = "forbid"


class RecommenderMeta(ABCMeta):
    recommender_name_vs_recommender_class: Dict[str, "RecommenderMeta"] = {}

    @no_type_check
    def __new__(
        mcs,
        name,
        bases,
        namespace,
        register_class: bool = True,
        **kwargs,
    ):

        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        if register_class:
            mcs.recommender_name_vs_recommender_class[name] = cls
        return cls


class BaseRecommender(object, metaclass=RecommenderMeta):
    """The base class for all (hot) recommenders.

    Args:
        X_train_all (csr_matrix|csc_matrix|np.ndarray): user/item interaction matrix.
            each row correspods to a user's interaction with items.
    """

    config_class: Type[RecommenderConfig]
    default_tune_range: List[ParameterRange]

    X_train_all: sps.csr_matrix
    """The matrix to feed into recommender."""

    def __init__(self, X_train_all: InteractionMatrix, **kwargs: Any) -> None:
        self.X_train_all: sps.csr_matrix = sps.csr_matrix(X_train_all).astype(
            np.float64
        )

        self.n_users: int = self.X_train_all.shape[0]
        self.n_items: int = self.X_train_all.shape[1]
        self.X_train_all.sort_indices()

        # this will store configurable parameters learnt during the training,
        # e.g., the epoch with the best validation score.
        self.learnt_config: Dict[str, Any] = dict()

    @classmethod
    def from_config(
        cls: Type[R],
        X_train_all: InteractionMatrix,
        config: RecommenderConfig,
    ) -> R:
        if not isinstance(config, cls.config_class):
            raise ValueError(
                f"Different config has been given. config must be {cls.config_class}"
            )
        return cls(X_train_all, **config.dict())

    def learn(self: R) -> R:
        """Learns and returns itself.

        Returns:
            The model after fitting process.
        """
        self._learn()
        return self

    @abstractmethod
    def _learn(self) -> None:
        raise NotImplementedError("_learn must be implemented.")

    def learn_with_optimizer(
        self,
        evaluator: Optional["evaluation.Evaluator"],
        trial: Optional["Trial"],
        max_epoch: int = 128,
        validate_epoch: int = 5,
        score_degradation_max: int = 5,
    ) -> None:
        r"""Learning procedures with early stopping and pruning.

        Args:
            evaluator : The evaluator to measure the score.
            trial : The current optuna trial under the study (if any.)
            max_epoch:
                Maximal number of epochs.
                If iterative learning procedure is not available, this parameter will be ignored.
                Defaults to 128.
            validate_epoch:
                The frequency of validation score measurement.
                If iterative learning procedure is not available, this parameter will be ignored.
                Defaults to 5.
            validate_epoch:
                The frequency of validation score measurement.
                If iterative learning procedure is not available, this parameter will be ignored.
                Defaults to 5.
            score_degradation_max:
                Maximal number of allowed score degradation.
                If iterative learning procedure is not available, this parameter will be ignored.
                Defaults to 5.
        """
        # by default, evaluator & trial does not play any role.
        self.learn()

    @classmethod
    def default_suggest_parameter(
        cls, trial: "Trial", fixed_params: Dict[str, Any]
    ) -> Dict[str, Any]:
        return {
            s.name: s.suggest(trial)
            for s in cls.default_tune_range
            if s.name not in fixed_params.keys()
        }

    @classmethod
    def tune(
        cls,
        data: Union[InteractionMatrix, None],
        evaluator: "evaluation.Evaluator",
        n_trials: int = 20,
        timeout: Optional[int] = None,
        data_suggest_function: Optional[Callable[["Trial"], InteractionMatrix]] = None,
        parameter_suggest_function: Optional[ParameterSuggestFunctionType] = None,
        fixed_params: Dict[str, Any] = dict(),
        random_seed: Optional[int] = None,
        prunning_n_startup_trials: int = 10,
        max_epoch: int = 128,
        validate_epoch: int = 5,
        score_degradation_max: int = 5,
        logger: Optional[logging.Logger] = None,
    ) -> Tuple[Dict[str, Any], pd.DataFrame]:
        r"""Perform the optimization step.
        `optuna.Study` object is created inside this function.

        Args:
            data:
                The training data.
                You can also provide tunable parameter dependent training data by providing `data_suggest_function`.
                In that case, data must be `None`.
            evaluator:
                The validation evaluator that measures the performance of the recommenders.
            n_trials:
                The number of expected trials (including pruned ones). Defaults to 20.
            timeout:
                If set to some value (in seconds), the study will exit after that time period.
                Note that the running trials is not interrupted, though. Defaults to `None`.
            data_suggest_function:
                If not `None`, this must be a function which takes `optuna.Trial` as its argument and returns training data. Defaults to `None`.
            parameter_suggest_function:
                If not `None`, this must be a function which takes `optuna.Trial` as its argument and returns `Dict[str, Any]` (i.e., some keyword arguments of the recommender class).
                If `None`, `cls.default_suggest_parameter` will be used for the parameter suggestion.
                Defaults to `None`.
            fixed_params:
                Fixed parameters passed to recommenders during the optimization procedure.
                This will replace the suggested parameter (either by `cls.default_suggest_parameter` or `parameter_suggest_function`).
                Defaults to `dict()`.
            random_seed:
                The random seed to control `optuna.samplers.TPESampler`. Defaults to `None`.
            prunning_n_startup_trials:
                `n_startup_trials` argument passed to the constructor of `optuna.pruners.MedianPruner`.
            max_epoch:
                The maximal number of epochs for the training.
                If iterative learning procedure is not available, this parameter will be ignored.
            validate_epoch (int, optional):
                The frequency of validation score measurement.
                If iterative learning procedure is not available, this parameter will be ignored.
                Defaults to 5.
            score_degradation_max (int, optional):
                Maximal number of allowed score degradation.
                If iterative learning procedure is not available, this parameter will be ignored.
                Defaults to 5. Defaults to 5.
        Returns:
            A tuple that consists of

                1. A dict containing the best paramaters.
                   This dict can be passed to the recommender as ``**kwargs``.
                2. A ``pandas.DataFrame`` that contains the history of optimization.
        """

        from optuna import create_study, pruners, samplers

        study = create_study(
            sampler=samplers.TPESampler(seed=random_seed),
            pruner=pruners.MedianPruner(n_startup_trials=prunning_n_startup_trials),
        )
        return cls.tune_with_study(
            study,
            data=data,
            evaluator=evaluator,
            n_trials=n_trials,
            timeout=timeout,
            data_suggest_function=data_suggest_function,
            parameter_suggest_function=parameter_suggest_function,
            fixed_params=fixed_params,
            max_epoch=max_epoch,
            validate_epoch=validate_epoch,
            score_degradation_max=score_degradation_max,
            logger=logger,
        )

    @classmethod
    def tune_with_study(
        cls,
        study: "Study",
        data: Union[InteractionMatrix, None],
        evaluator: "evaluation.Evaluator",
        n_trials: int = 20,
        timeout: Optional[int] = None,
        data_suggest_function: Optional[Callable[["Trial"], InteractionMatrix]] = None,
        parameter_suggest_function: Optional[ParameterSuggestFunctionType] = None,
        fixed_params: Dict[str, Any] = dict(),
        max_epoch: int = 128,
        validate_epoch: int = 5,
        score_degradation_max: int = 5,
        logger: Optional[logging.Logger] = None,
    ) -> Tuple[Dict[str, Any], pd.DataFrame]:
        from ..optimization.optimizer import Optimizer

        if data is None:
            if data_suggest_function is None:
                raise ValueError(
                    "Either `data` or `data_sugget_function` must be provided."
                )
            _data_suggest_function = data_suggest_function
        else:

            def _data_suggest_function(trial: "Trial") -> InteractionMatrix:
                return data

        if parameter_suggest_function is not None:
            _parameter_suggest_function = parameter_suggest_function
        else:

            def _parameter_suggest_function(trial: "Trial") -> Dict[str, Any]:
                return cls.default_suggest_parameter(trial, fixed_params)

        optim = Optimizer(
            _data_suggest_function,
            _parameter_suggest_function,
            fixed_params,
            evaluator,
            max_epoch=max_epoch,
            validate_epoch=validate_epoch,
            score_degradation_max=score_degradation_max,
            logger=logger,
        )
        return optim.optimize_with_study(study, cls, n_trials=n_trials, timeout=timeout)

    @abstractmethod
    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        """Compute the item recommendation score for a subset of users.

        Args:
            user_indices : The index defines the subset of users.

        Returns:
            The item scores. Its shape will be (len(user_indices), self.n_items)
        """
        raise NotImplementedError("get_score must be implemented")  # pragma: no cover

    def get_score_block(self, begin: int, end: int) -> DenseScoreArray:
        """Compute the score for a block of the users.

        Args:
            begin (int): where the evaluated user block begins.
            end (int): where the evaluated user block ends.

        Returns:
            The item scores. Its shape will be (end - begin, self.n_items)
        """
        raise NotImplementedError("get_score_block not implemented!")

    def get_score_remove_seen(self, user_indices: UserIndexArray) -> DenseScoreArray:
        """Compute the item score and mask the item in the training set. Masked items will have the score -inf.

        Args:
            user_indices : Specifies the subset of users.

        Returns:
            The masked item scores. Its shape will be (len(user_indices), self.n_items)
        """
        scores = self.get_score(user_indices)
        if sps.issparse(scores):
            scores = sps.csr_matrix(scores).toarray()
        m = self.X_train_all[user_indices].tocsr()
        scores[m.nonzero()] = -np.inf
        return scores

    def get_score_remove_seen_block(self, begin: int, end: int) -> DenseScoreArray:
        """Compute the score for a block of the users, and mask the items in the training set. Masked items will have the score -inf.

        Args:
            begin (int): where the evaluated user block begins.
            end (int): where the evaluated user block ends.

        Returns:
            The masked item scores. Its shape will be (end - begin, self.n_items)
        """
        scores = _sparse_to_array(self.get_score_block(begin, end))
        m = self.X_train_all[begin:end]
        scores[m.nonzero()] = -np.inf
        return scores

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        """Compute the item recommendation score for unseen users whose profiles are given as another user-item relation matrix.

        Args:
            X : The profile user-item relation matrix for unseen users.
                Its number of rows is arbitrary, but the number of columns must be self.n_items.

        Returns:
            Computed item scores for users. Its shape is equal to X.
        """
        raise NotImplementedError(
            f"get_score_cold_user is not implemented for {self.__class__.__name__}!"
        )  # pragma: no cover

    def get_score_cold_user_remove_seen(self, X: InteractionMatrix) -> DenseScoreArray:
        """Compute the item recommendation score for unseen users whose profiles are given as another user-item relation matrix. The score will then be masked by the input.

        Args:
            X : The profile user-item relation matrix for unseen users.
                Its number of rows is arbitrary, but the number of columns must be self.n_items.

        Returns:
            Computed & masked item scores for users. Its shape is equal to X.
        """
        score = self.get_score_cold_user(X)
        score[X.nonzero()] = -np.inf
        return score


class BaseSimilarityRecommender(BaseRecommender):
    """The computed item-item similarity. Might not be initialized before `learn()` is called."""

    _W: Optional[Union[sps.csr_matrix, sps.csc_matrix, np.ndarray]]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._W = None

    @property
    def W(self) -> Union[sps.csr_matrix, sps.csc_matrix, np.ndarray]:
        """The computed item-item similarity weight matrix."""
        if self._W is None:
            raise RuntimeError("W fetched before fit.")
        return self._W

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        return _sparse_to_array(self.X_train_all[user_indices].dot(self.W))

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        return _sparse_to_array(X.dot(self.W))

    def get_score_block(self, begin: int, end: int) -> DenseScoreArray:
        return _sparse_to_array(self.X_train_all[begin:end].dot(self.W))


class BaseUserSimilarityRecommender(BaseRecommender):
    """The computed user-user similarity. Might not be initialized before `learn()` is called."""

    U_: Optional[Union[sps.csr_matrix, sps.csc_matrix, np.ndarray]]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._X_csc: sps.csc_matrix = self.X_train_all.tocsc()
        self.U_ = None

    @property
    def U(self) -> Union[sps.csr_matrix, sps.csc_matrix, np.ndarray]:
        """The computed user-user similarity weight matrix."""
        if self.U_ is None:
            raise RuntimeError("W fetched before fit.")
        return self.U_

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        return _sparse_to_array(self.U[user_indices].dot(self._X_csc).toarray())

    def get_score_block(self, begin: int, end: int) -> DenseScoreArray:
        return _sparse_to_array(self.U[begin:end].dot(self._X_csc))


class BaseRecommenderWithUserEmbedding:
    r"""Defines a recommender with user embedding (e.g., matrix factorization.)."""

    @abstractmethod
    def get_user_embedding(
        self,
    ) -> DenseMatrix:
        """Get user embedding vectors.

        Returns:
            The latent vector representation of users.
            Its number of rows is equal to the number of the users.
        """

    @abstractmethod
    def get_score_from_user_embedding(
        self, user_embedding: DenseMatrix
    ) -> DenseScoreArray:
        """Compute the item score from user embedding. Mainly used for cold-start scenario.

        Args:
            user_embedding : Latent user representation obtained elsewhere.

        Returns:
            DenseScoreArray: The score array. Its shape will be ``(user_embedding.shape[0], self.n_items)``
        """
        raise NotImplementedError("get_score_from_item_embedding must be implemtented.")


class BaseRecommenderWithItemEmbedding:
    """Defines a recommender with item embedding (e.g., matrix factorization.)."""

    @abstractmethod
    def get_item_embedding(
        self,
    ) -> DenseMatrix:
        """Get item embedding vectors.

        Returns:
            The latent vector representation of items.
            Its number of rows is equal to the number of the items.
        """
        raise NotImplementedError(
            "get_item_embedding must be implemented."
        )  # pragma: no cover

    @abstractmethod
    def get_score_from_item_embedding(
        self, user_indices: UserIndexArray, item_embedding: DenseMatrix
    ) -> DenseScoreArray:
        raise NotImplementedError("get_score_from_item_embedding must be implemented.")


def get_recommender_class(recommender_name: str) -> Type[BaseRecommender]:

    r"""Get recommender class from its class name.

    Args:
        recommender_name: The class name of the recommender.

    Returns:
        The recommender class with its class name being `recommender_name`.
    """
    result: Type[
        BaseRecommender
    ] = RecommenderMeta.recommender_name_vs_recommender_class[recommender_name]
    return result



import re
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, List

if TYPE_CHECKING:
    from optuna import Trial

NAME_CHECKER = re.compile(r"^([a-zA-Z\d]+[\-_]*)+$")


def is_valid_param_name(name: str) -> bool:
    if NAME_CHECKER.match(name) is None:
        return False
    return True


class ParameterRange(metaclass=ABCMeta):
    r"""The class to define default parameter tuning range.

    Args:
        name (str): The name of the parameter to be tuned.
    """

    def __init__(self, name: str):
        if not is_valid_param_name(name):
            raise ValueError(
                rf""""{name}" is  not a valid parameter name. It should match r"^([a-zA-Z\d]+[\-_]*)+$"""
            )
        self.name = name

    @abstractmethod
    def suggest(self, trial: "Trial", prefix: str = "") -> Any:
        raise NotImplementedError('"suggest" must be implemented.')  # pragma: no cover


class UniformFloatRange(ParameterRange):
    def __init__(self, name: str, low: float, high: float):
        super().__init__(name)
        if low > high:
            raise ValueError("Got low > high.")

        self.low = low
        self.high = high

    def suggest(self, trial: "Trial", prefix: str = "") -> Any:
        return trial.suggest_float(prefix + self.name, self.low, self.high)


class LogUniformFloatRange(ParameterRange):
    def __init__(self, name: str, low: float, high: float):
        super().__init__(name)
        if low > high:
            raise ValueError("Got low > high.")

        self.low = low
        self.high = high

    def suggest(self, trial: "Trial", prefix: str = "") -> Any:
        return trial.suggest_float(prefix + self.name, self.low, self.high, log=True)


class UniformIntegerRange(ParameterRange):
    def __init__(self, name: str, low: int, high: int, step: int = 1):
        super().__init__(name)
        if low > high:
            raise ValueError("Got low > high.")

        self.low = low
        self.high = high
        self.step = step

    def suggest(self, trial: "Trial", prefix: str = "") -> Any:
        return trial.suggest_int(
            prefix + self.name, self.low, self.high, step=self.step
        )


class LogUniformIntegerRange(ParameterRange):
    def __init__(self, name: str, low: int, high: int):
        super().__init__(name)
        if low > high:
            raise ValueError("Got low > high.")

        self.low = low
        self.high = high

    def suggest(self, trial: "Trial", prefix: str = "") -> Any:
        return trial.suggest_int(prefix + self.name, self.low, self.high, log=True)


class CategoricalRange(ParameterRange):
    def __init__(self, name: str, choices: List[Any]):
        super().__init__(name)
        self.choices = choices

    def suggest(self, trial: "Trial", prefix: str = "") -> Any:
        return trial.suggest_categorical(prefix + self.name, self.choices)


default_tune_range_knn = [
    UniformIntegerRange("top_k", 4, 1000),
    UniformFloatRange("shrinkage", 0, 1000),
]

default_tune_range_knn_with_weighting = [
    UniformIntegerRange("top_k", 4, 1000),
    UniformFloatRange("shrinkage", 0, 1000),
    CategoricalRange("feature_weighting", ["NONE", "TF_IDF", "BM_25"]),
]



from typing import Any, Dict, List, Optional

import pandas as pd

from .._threading import get_n_threads
from ._core import evaluate_list_vs_list


def evaluate_recommendation_df(
    df_recommendation: pd.DataFrame,
    df_ground_truth: pd.DataFrame,
    user_column: str,
    item_column: str,
    n_threads: Optional[int] = None,
) -> Dict[str, float]:
    r"""Measure accuracy/diversity metrics by providing recommendation result & ground truth as  `pd.DataFrame`s.

    Args:
        df_recommendation:
            The dataframe for the recommendation results.
            In this dataframe, we assume that the more important the recommended item, the earlier it will appear.
        df_ground_truth:
            The dataframe for the groud truth.
        user_column (str):
            Specifies which column reporesents the users' id.
        item_column (str):
            Specifies which column reporesents the items' id.
        n_threads:
            Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if the variable is not set, it will be set to ``os.cpu_count()``. Defaults to None.

    Returns:
        Resulting accuracy/diversity metrics.
    """
    df_recommendation = df_recommendation[
        df_recommendation[user_column].isin(df_ground_truth[user_column])
    ].copy()

    df_ground_truth = df_ground_truth[
        df_ground_truth[user_column].isin(df_recommendation[user_column])
    ].copy()

    user_ids = sorted(list(set(df_recommendation[user_column])))
    user_id_to_index: Dict[Any, int] = {uid: i for i, uid in enumerate(user_ids)}
    item_ids = sorted(
        list(set(df_recommendation[item_column]) | set(df_ground_truth[item_column]))
    )
    item_id_to_index: Dict[Any, int] = {iid: i for i, iid in enumerate(item_ids)}
    user_index_to_recommendation: Dict[int, List[int]] = {}
    user_index_to_ground_truth: Dict[int, List[int]] = {}
    for rec_row in df_recommendation.itertuples():
        uid = getattr(rec_row, user_column)
        iid = getattr(rec_row, item_column)
        u_index = user_id_to_index[uid]
        i_index = item_id_to_index[iid]
        user_index_to_recommendation.setdefault(u_index, []).append(i_index)

    for gt_row in df_ground_truth.itertuples():
        uid = getattr(gt_row, user_column)
        iid = getattr(gt_row, item_column)
        u_index = user_id_to_index[uid]
        i_index = item_id_to_index[iid]
        user_index_to_ground_truth.setdefault(u_index, []).append(i_index)
    recommendation_list: List[List[int]] = []
    ground_truth_list: List[List[int]] = []
    for uindex, recommendation in user_index_to_recommendation.items():
        recommendation_list.append(recommendation)
        gts = user_index_to_ground_truth.get(uindex, [])
        ground_truth_list.append(gts)

    return evaluate_list_vs_list(
        recommendation_list,
        ground_truth_list,
        len(item_id_to_index),
        get_n_threads(n_threads),
    ).as_dict()



from uuid import uuid1

import numpy as np
import pandas as pd
import pytest
from scipy import sparse as sps

from irspack.evaluation.evaluate_df_to_df import evaluate_recommendation_df
from irspack.evaluation.evaluator import Evaluator
from irspack.recommenders.ials import IALSRecommender
from irspack.utils import IDMapper, rowwise_train_test_split


def test_matching() -> None:
    n_users = 30
    n_items = 20
    cutoff = 3
    user_column = "UserId"
    item_column = "ItemId"
    _, X = rowwise_train_test_split(
        sps.csr_matrix(np.ones((n_users, n_items), dtype=np.float64)),
        n_test=n_users // 2,
        random_state=0,
    )
    assert np.all(X.data == 1)
    X_train, X_test = rowwise_train_test_split(X, random_state=0, n_test=3)
    assert X_train.multiply(X_test).getnnz() == 0
    del X
    user_ids = [str(uuid1()) for _ in range(n_users)]
    item_ids = [str(uuid1()) for _ in range(n_items)]

    def _sps_to_df(X: sps.csr_matrix) -> pd.DataFrame:
        X_as_list = []
        for row, col in zip(*X.nonzero()):
            X_as_list.append({user_column: user_ids[row], item_column: item_ids[col]})
        return pd.DataFrame(X_as_list)

    X_train_as_df = _sps_to_df(X_train)
    X_test_as_df = _sps_to_df(X_test)

    rec = IALSRecommender(X_train, n_components=1, train_epochs=0).learn()
    id_mapper = IDMapper(user_ids, item_ids)

    rec_list = id_mapper.recommend_for_known_user_batch(rec, user_ids, cutoff=cutoff)
    recommemdation_list = []
    for uid, recommend_per_user in zip(user_ids, rec_list):
        for iid, _ in recommend_per_user:
            recommemdation_list.append({user_column: uid, item_column: iid})
    recommendation_df = pd.DataFrame(recommemdation_list)
    assert recommendation_df.merge(X_train_as_df).shape[0] == 0
    evaluator = Evaluator(X_test, cutoff=cutoff, n_threads=1)
    score_using_evaluator = evaluator.get_score(rec)
    score_using_df_vs_df = evaluate_recommendation_df(
        recommendation_df, X_test_as_df, user_column, item_column, n_threads=1
    )
    assert score_using_evaluator["ndcg"] == pytest.approx(score_using_df_vs_df["ndcg"])
    assert score_using_evaluator["map"] == pytest.approx(score_using_df_vs_df["map"])
    assert score_using_evaluator["hit"] == pytest.approx(score_using_df_vs_df["hit"])
    assert score_using_evaluator["recall"] == pytest.approx(
        score_using_df_vs_df["recall"]
    )
    assert score_using_evaluator["precision"] == pytest.approx(
        score_using_df_vs_df["precision"]
    )
    assert score_using_evaluator["entropy"] == pytest.approx(
        score_using_df_vs_df["entropy"]
    )
    entropy = score_using_evaluator["entropy"]
    item_freq_cnt = (
        recommendation_df[
            recommendation_df[user_column].isin(X_test_as_df[user_column])
        ][item_column]
        .value_counts()
        .values
    )
    item_freq_prob = item_freq_cnt.astype(float) / float(item_freq_cnt.sum())
    entropy_gt = -np.log(item_freq_prob).dot(item_freq_prob)
    assert entropy == pytest.approx(entropy_gt)



m: int
n: int
from numpy import float32

"""irspack's core module for "IALSRecommender".
Built to use
	SSE, SSE2"""
from __future__ import annotations
import irspack.recommenders._ials
import typing
import numpy
import scipy.sparse

_Shape = typing.Tuple[int, ...]

__all__ = [
    "CG",
    "CHOLESKY",
    "IALSModelConfig",
    "IALSModelConfigBuilder",
    "IALSPP",
    "IALSSolverConfig",
    "IALSSolverConfigBuilder",
    "IALSTrainer",
    "LossType",
    "ORIGINAL",
    "SolverType",
]

class IALSModelConfig:
    def __getstate__(self) -> tuple: ...
    def __init__(
        self,
        arg0: int,
        arg1: float,
        arg2: float,
        arg3: float,
        arg4: float,
        arg5: int,
        arg6: LossType,
    ) -> None: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    pass

class IALSModelConfigBuilder:
    def __init__(self) -> None: ...
    def build(self) -> IALSModelConfig: ...
    def set_K(self, arg0: int) -> IALSModelConfigBuilder: ...
    def set_alpha0(self, arg0: float) -> IALSModelConfigBuilder: ...
    def set_init_stdev(self, arg0: float) -> IALSModelConfigBuilder: ...
    def set_loss_type(self, arg0: LossType) -> IALSModelConfigBuilder: ...
    def set_nu(self, arg0: float) -> IALSModelConfigBuilder: ...
    def set_random_seed(self, arg0: int) -> IALSModelConfigBuilder: ...
    def set_reg(self, arg0: float) -> IALSModelConfigBuilder: ...
    pass

class IALSSolverConfig:
    def __getstate__(self) -> tuple: ...
    def __init__(
        self, arg0: int, arg1: SolverType, arg2: int, arg3: int, arg4: int
    ) -> None: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    pass

class IALSSolverConfigBuilder:
    def __init__(self) -> None: ...
    def build(self) -> IALSSolverConfig: ...
    def set_ialspp_iteration(self, arg0: int) -> IALSSolverConfigBuilder: ...
    def set_ialspp_subspace_dimension(self, arg0: int) -> IALSSolverConfigBuilder: ...
    def set_max_cg_steps(self, arg0: int) -> IALSSolverConfigBuilder: ...
    def set_n_threads(self, arg0: int) -> IALSSolverConfigBuilder: ...
    def set_solver_type(self, arg0: SolverType) -> IALSSolverConfigBuilder: ...
    pass

class IALSTrainer:
    def __getstate__(self) -> tuple: ...
    def __init__(
        self, arg0: IALSModelConfig, arg1: scipy.sparse.csr_matrix[numpy.float32]
    ) -> None: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    def compute_loss(self, arg0: IALSSolverConfig) -> float: ...
    def step(self, arg0: IALSSolverConfig) -> None: ...
    def transform_item(
        self, arg0: scipy.sparse.csr_matrix[numpy.float32], arg1: IALSSolverConfig
    ) -> numpy.ndarray[numpy.float32, _Shape[m, n]]: ...
    def transform_user(
        self, arg0: scipy.sparse.csr_matrix[numpy.float32], arg1: IALSSolverConfig
    ) -> numpy.ndarray[numpy.float32, _Shape[m, n]]: ...
    def user_scores(
        self, arg0: int, arg1: int, arg2: IALSSolverConfig
    ) -> numpy.ndarray[numpy.float32, _Shape[m, n]]: ...
    @property
    def item(self) -> numpy.ndarray[numpy.float32, _Shape[m, n]]:
        """
        :type: numpy.ndarray[numpy.float32, _Shape[m, n]]
        """
    @item.setter
    def item(self, arg0: numpy.ndarray[numpy.float32, _Shape[m, n]]) -> None:
        pass
    @property
    def user(self) -> numpy.ndarray[numpy.float32, _Shape[m, n]]:
        """
        :type: numpy.ndarray[numpy.float32, _Shape[m, n]]
        """
    @user.setter
    def user(self, arg0: numpy.ndarray[numpy.float32, _Shape[m, n]]) -> None:
        pass
    pass

class LossType:
    """
    Members:

      ORIGINAL

      IALSPP
    """

    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str:
        """
        :type: str
        """
    @property
    def value(self) -> int:
        """
        :type: int
        """
    IALSPP: irspack.recommenders._ials.LossType  # value = <LossType.IALSPP: 1>
    ORIGINAL: irspack.recommenders._ials.LossType  # value = <LossType.ORIGINAL: 0>
    __members__: dict  # value = {'ORIGINAL': <LossType.ORIGINAL: 0>, 'IALSPP': <LossType.IALSPP: 1>}
    pass

class SolverType:
    """
    Members:

      CHOLESKY

      CG

      IALSPP
    """

    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str:
        """
        :type: str
        """
    @property
    def value(self) -> int:
        """
        :type: int
        """
    CG: irspack.recommenders._ials.SolverType  # value = <SolverType.CG: 1>
    CHOLESKY: irspack.recommenders._ials.SolverType  # value = <SolverType.CHOLESKY: 0>
    IALSPP: irspack.recommenders._ials.SolverType  # value = <SolverType.IALSPP: 2>
    __members__: dict  # value = {'CHOLESKY': <SolverType.CHOLESKY: 0>, 'CG': <SolverType.CG: 1>, 'IALSPP': <SolverType.IALSPP: 2>}
    pass

CG: irspack.recommenders._ials.SolverType  # value = <SolverType.CG: 1>
CHOLESKY: irspack.recommenders._ials.SolverType  # value = <SolverType.CHOLESKY: 0>
IALSPP: irspack.recommenders._ials.SolverType  # value = <SolverType.IALSPP: 2>
ORIGINAL: irspack.recommenders._ials.LossType  # value = <LossType.ORIGINAL: 0>



import random
import uuid
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pytest
import scipy.sparse as sps

from irspack.definitions import DenseScoreArray, InteractionMatrix
from irspack.recommenders import BaseRecommender
from irspack.utils.id_mapping import IDMapper, ItemIDMapper


class MockRecommender(BaseRecommender):
    def __init__(self, X_all: sps.csr_matrix, scores: np.ndarray) -> None:
        super().__init__(X_all)
        self.scores = scores

    def get_score(self, user_indices: np.ndarray) -> np.ndarray:
        return self.scores[user_indices]

    def _learn(self) -> None:
        pass

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        score = np.exp(X.toarray())
        score /= score.sum(axis=1)[:, None]
        return score


def check_descending(id_score: List[Tuple[Any, float]]) -> None:
    for i, (_, score) in enumerate(id_score):
        if i > 0:
            if score > id_score[i - 1][1]:
                raise RuntimeError("not descending!")


def test_check_descending_1() -> None:
    check_descending([("1", 100), ("2", 99), ("100", 99), ("3", -1000)])
    with pytest.raises(RuntimeError):
        check_descending([("1", 100), ("2", 99), ("3", -1000), ("", -999)])


@pytest.mark.parametrize("dtype", ["float32", "float64", "float16"])
def test_basic_usecase(dtype: str) -> None:
    RNS = np.random.RandomState(0)
    n_users = 31
    n_items = 42
    user_ids = [str(uuid.uuid4()) for _ in range(n_users)]
    item_ids = [str(uuid.uuid4()) for _ in range(n_items)]
    item_id_set = set(item_ids)
    score = RNS.randn(n_users, n_items).astype(dtype)
    X = sps.csr_matrix((score + RNS.randn(*score.shape)) > 0).astype(np.float64)
    rec = MockRecommender(X, score).learn()

    with pytest.raises(ValueError):
        id_mapper_extra_item_ids = IDMapper(user_ids, item_ids + [str(uuid.uuid4())])
        id_mapper_extra_item_ids.recommend_for_known_user_id(rec, user_ids[0])

    with pytest.raises(ValueError):
        id_mapper_extra_user_ids = IDMapper(user_ids + [str(uuid.uuid4())], item_ids)
        id_mapper_extra_user_ids.recommend_for_known_user_id(rec, user_ids[0])

    id_mapper = IDMapper(user_ids, item_ids)

    with pytest.raises(RuntimeError):
        id_mapper.recommend_for_known_user_id(rec, str(uuid.uuid4()))

    known_user_results_individual = []
    for i, uid in enumerate(user_ids):
        nonzero_items = [item_ids[j] for j in X[i].nonzero()[1]]
        recommended = id_mapper.recommend_for_known_user_id(rec, uid, cutoff=n_items)
        check_descending(recommended)
        recommended_ids = {rec[0] for rec in recommended}
        assert len(recommended_ids.intersection(nonzero_items)) == 0
        assert len(recommended_ids.union(nonzero_items)) == n_items
        cutoff = n_items // 2
        recommendation_with_cutoff = id_mapper.recommend_for_known_user_id(
            rec, uid, cutoff=cutoff
        )
        known_user_results_individual.append(recommendation_with_cutoff)
        check_descending(recommendation_with_cutoff)
        assert len(recommendation_with_cutoff) <= cutoff

        forbbiden_item = list(
            set(
                RNS.choice(
                    list(item_id_set.difference(nonzero_items)),
                    replace=False,
                    size=(n_items - len(nonzero_items)) // 2,
                )
            )
        )

        recommended_with_restriction = id_mapper.recommend_for_known_user_id(
            rec, uid, cutoff=n_items, forbidden_item_ids=forbbiden_item
        )
        check_descending(recommended_with_restriction)
        for iid, score in recommended_with_restriction:
            assert iid not in forbbiden_item

        random_allowed_items = list(
            RNS.choice(
                list(item_id_set.difference(nonzero_items)),
                size=min(n_items - len(nonzero_items), n_items // 3),
            )
        )
        random_allowed_items += [str(uuid.uuid1())]
        with_allowed_item = id_mapper.recommend_for_known_user_id(
            rec,
            uid,
            cutoff=n_items,
            allowed_item_ids=random_allowed_items,
        )
        check_descending(with_allowed_item)
        for id_, _ in with_allowed_item:
            assert id_ in random_allowed_items

        allowed_only_forbidden = id_mapper.recommend_for_known_user_id(
            rec,
            uid,
            cutoff=n_items,
            forbidden_item_ids=forbbiden_item,
            allowed_item_ids=forbbiden_item,
        )
        assert not allowed_only_forbidden

        recommendations_for_coldstart = id_mapper.recommend_for_new_user(
            rec, nonzero_items, cutoff=n_items
        )
        coldstart_rec_results = {_[0] for _ in recommendations_for_coldstart}
        assert len(coldstart_rec_results.intersection(nonzero_items)) == 0
        assert len(coldstart_rec_results.union(nonzero_items)) == n_items

    if dtype == "float16":
        with pytest.raises(ValueError):
            known_user_results_batch = id_mapper.recommend_for_known_user_batch(
                rec, user_ids, cutoff=n_items
            )
        return
    known_user_results_batch = id_mapper.recommend_for_known_user_batch(
        rec, user_ids, cutoff=n_items
    )
    assert len(known_user_results_batch) == len(known_user_results_individual)
    for batch_res, indiv_res in zip(
        known_user_results_batch, known_user_results_individual
    ):
        for (id_batch, score_batch), (id_indiv, score_indiv) in zip(
            batch_res, indiv_res
        ):
            assert id_batch == id_indiv
            assert score_batch == pytest.approx(score_indiv)

    nonzero_batch: List[List[str]] = []
    for i, _ in enumerate(user_ids):
        nonzero_items = [item_ids[j] for j in X[i].nonzero()[1]]
        nonzero_batch.append(nonzero_items)

    batch_result_non_masked = id_mapper.recommend_for_new_user_batch(
        rec, nonzero_batch, cutoff=n_items, n_threads=1
    )

    assert len(batch_result_non_masked) == n_users
    for recommendation_using_batch, uid, nonzero_items in zip(
        batch_result_non_masked, user_ids, nonzero_batch
    ):
        check_descending(recommendation_using_batch)
        recommended_ids = {rec[0] for rec in recommendation_using_batch}
        assert len(recommended_ids.intersection(nonzero_items)) == 0
        assert len(recommended_ids.union(nonzero_items)) == n_items

    batch_result_non_masked_cutoff_3 = id_mapper.recommend_for_new_user_batch(
        rec, nonzero_batch, cutoff=3, n_threads=2
    )

    assert len(batch_result_non_masked_cutoff_3) == n_users
    for recommended_using_batch_cutoff_3, uid, nonzero_items in zip(
        batch_result_non_masked_cutoff_3, user_ids, nonzero_batch
    ):
        assert len(recommended_using_batch_cutoff_3) == 3
        check_descending(recommended_using_batch_cutoff_3)
        recommended_ids = {rec[0] for rec in recommended_using_batch_cutoff_3}
        for rid in recommended_ids:
            assert rid not in nonzero_items

    batch_result_masked = id_mapper.recommend_for_new_user_batch(
        rec,
        nonzero_batch,
        cutoff=n_items,
        n_threads=1,
    )
    assert len(batch_result_masked) == n_users
    for recommended_using_batch_masked, uid, nonzero_items in zip(
        batch_result_masked, user_ids, nonzero_batch
    ):
        check_descending(recommended_using_batch_masked)
        recommended_ids = {rec[0] for rec in recommended_using_batch_masked}
        assert len(recommended_ids.intersection(nonzero_items)) == 0

    nonzero_batch_dict: List[Dict[str, float]] = []
    for i, _ in enumerate(user_ids):
        profile = {item_ids[j]: 2.0 for j in X[i].nonzero()[1]}
        nonzero_batch_dict.append(profile)
    batch_result_dict = id_mapper.recommend_for_new_user_batch(
        rec, nonzero_batch_dict, cutoff=n_items, n_threads=1
    )
    for i, result in enumerate(batch_result_dict):
        nnz = len(X[i].nonzero()[1])
        softmax_denom = X.shape[1] - nnz + np.exp(2) * nnz
        for _, score in result:
            assert score == pytest.approx(1 / softmax_denom)


def test_item_id_mapper_per_user_allowed_item() -> None:
    ids = [(i, f"{i}") for i in range(1, 11)]
    np.random.shuffle(ids)
    id_mapepr = ItemIDMapper(ids)
    N_users = 10
    score = np.random.random((N_users, len(ids)))
    per_user_allowed_item_ids = [
        list(set([random.choice(ids) for _ in range(random.randint(2, 5))]))
        for __ in range(N_users)
    ]
    recommendation = id_mapepr.score_to_recommended_items_batch(
        score, cutoff=5, per_user_allowed_item_ids=per_user_allowed_item_ids
    )
    assert len(recommendation) == score.shape[0]
    for i, (rec, allowed_item_ids_for_user) in enumerate(
        zip(recommendation, per_user_allowed_item_ids)
    ):
        previous_score: Optional[float] = None
        for ((id_int, id_str), score_value) in rec:
            assert (id_int, id_str) in allowed_item_ids_for_user
            item_index = ids.index((id_int, f"{id_int}"))
            assert score[i, item_index] == pytest.approx(score_value)
            if previous_score is not None:
                assert score_value <= previous_score
            previous_score = score_value


def test_item_id_mapper_uniform_allowed_item() -> None:
    ids = [(i, f"{i}") for i in range(1, 11)]
    np.random.shuffle(ids)
    id_mapepr = ItemIDMapper(ids)
    N_users = 10
    score = np.random.random((N_users, len(ids)))

    allowed_items_uniform = list(set([random.choice(ids) for _ in range(5)]))
    assert allowed_items_uniform
    recommendation = id_mapepr.score_to_recommended_items_batch(
        score, cutoff=5, allowed_item_ids=allowed_items_uniform
    )
    assert len(recommendation) == score.shape[0]
    for i, rec in enumerate(recommendation):
        previous_score: Optional[float] = None
        for ((id_int, id_str), score_value) in rec:
            assert (id_int, id_str) in allowed_items_uniform
            item_index = ids.index((id_int, f"{id_int}"))
            assert score[i, item_index] == pytest.approx(score_value)
            if previous_score is not None:
                assert score_value <= previous_score
            previous_score = score_value


def test_item_id_mapper_per_user_allowed_item_forbidden_items() -> None:
    ids: List[int] = list(range(10))
    per_user_allowed_item_ids = [
        [0, 1, 2, 1048576],
        [0, 1],
        [1, 2, -1],
    ]
    forbidden_items = [[0, 1, 2], [0], [1]]
    # so
    # Nothing can be recommended for user 0
    # 1 is the only choice for user 1
    # 2 is the only choice for user 2

    N_users = len(per_user_allowed_item_ids)
    assert len(forbidden_items) == N_users
    np.random.shuffle(ids)
    id_mapepr = ItemIDMapper(ids)
    score = np.random.random((len(forbidden_items), len(ids)))
    recommendation = id_mapepr.score_to_recommended_items_batch(
        score,
        cutoff=10,
        per_user_allowed_item_ids=per_user_allowed_item_ids,
        forbidden_item_ids=forbidden_items,
    )
    assert len(recommendation) == score.shape[0]

    user_0_recommendation = recommendation[0]
    assert not user_0_recommendation

    user_1_recommendation = recommendation[1]
    assert len(user_1_recommendation) == 1
    assert user_1_recommendation[0][0] == 1

    user_2_recommendation = recommendation[2]
    assert len(user_2_recommendation) == 1
    assert user_2_recommendation[0][0] == 2



import scipy.sparse as sps
from scipy.special import expit

from ..definitions import OptionalRandomState
from .random import convert_randomstate


def mf_example_data(
    n_users: int = 100,
    n_items: int = 100,
    n_components: int = 5,
    random_state: OptionalRandomState = None,
    density_target: float = 0.3,
) -> sps.csr_matrix:
    rns = convert_randomstate(random_state)
    user_factors = rns.randn(n_users, n_components) / (n_components**0.5)
    item_factors = rns.randn(n_items, n_components) / (n_components**0.5)
    eps = 1e-5
    bias_max = 100.0
    bias_min = -100.0
    bias_try = 0.0
    all_scores = user_factors.dot(item_factors.T)
    for _ in range(100):
        expected_density = expit(all_scores.ravel() + bias_try).mean()
        if expected_density > density_target:
            bias_max = bias_try
        else:
            bias_min = bias_try
        if (bias_max - bias_min) < eps:
            break
        bias_try = (bias_max + bias_min) / 2.0

    score = expit(user_factors.dot(item_factors.T) + bias_try)
    return sps.csr_matrix(rns.binomial(1, score))



import numpy as np
import scipy.sparse as sps

from irspack.recommenders.base import BaseRecommender


class MockRecommender(BaseRecommender, register_class=False):
    def __init__(self, X_all: sps.csr_matrix, scores: np.ndarray) -> None:
        super().__init__(X_all)
        assert X_all.shape == scores.shape
        self.scores = scores

    def get_score(self, user_indices: np.ndarray) -> np.ndarray:
        return self.scores[user_indices]

    def _learn(self) -> None:
        pass



from abc import ABC, abstractmethod
from io import BytesIO
from typing import IO, TYPE_CHECKING, Any, Optional, Type

from fastprogress import progress_bar

from ..definitions import InteractionMatrix
from .base import BaseRecommender, RecommenderConfig

if TYPE_CHECKING:
    from optuna import Trial

    from irspack import evaluation


class TrainerBase(ABC):
    """The trainer class for early-stoppable recommenders.
    The training logic of such recommenders (run-epoch, get-score) will be implemtented in the trainers."""

    def __init__(self, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def load_state(self, ifs: IO) -> None:
        """Load past state (with best validation score).

        Args:
            ifs (IO): The stream from which past state will be restored.
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def save_state(self, ofs: IO) -> None:
        """Save the current state into stream.

        Args:
            ofs (IO): the stream to which current state will be saved.
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def run_epoch(self) -> None:
        """Run single epoch."""
        raise NotImplementedError()  # pragma: no cover


class BaseEarlyStoppingRecommenderConfig(RecommenderConfig):
    train_epochs: int = 512


class BaseRecommenderWithEarlyStopping(BaseRecommender):
    """The base class for all the early-stoppable recommenders.

    Args:
        X_train_all:
            The train interaction matrix.
        train_epochs:
            The number of training epochs to run. Defaults to 128.
    """

    trainer_class: Type[TrainerBase]

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        train_epochs: int = 128,
        **kwargs: Any,
    ):

        super().__init__(X_train_all, **kwargs)
        self.train_epochs = train_epochs
        self.trainer: Optional[TrainerBase] = None
        self.best_state: Optional[bytes] = None

    @abstractmethod
    def _create_trainer(self) -> TrainerBase:
        pass  # pragma: no cover

    def start_learning(self) -> None:
        self.trainer = self._create_trainer()

    def run_epoch(self) -> None:
        if self.trainer is None:
            raise RuntimeError("'run_epoch' called before initializing the trainer.")
        self.trainer.run_epoch()

    def save_state(self) -> None:
        if self.trainer is None:
            raise RuntimeError("'save_state' called before initializing the trainer.")
        with BytesIO() as ofs:
            self.trainer.save_state(ofs)
            self.best_state = ofs.getvalue()

    def load_state(self) -> None:
        if self.trainer is None:
            raise RuntimeError("'load_state' called before initializing the trainer.")
        if self.best_state is None:
            raise RuntimeError("'load_state' called before achieving any results.")
        with BytesIO(self.best_state) as ifs:
            self.trainer.load_state(ifs)

    def _learn(self) -> None:
        self.learn_with_optimizer(None, None, max_epoch=self.train_epochs)

    def learn_with_optimizer(
        self,
        evaluator: Optional["evaluation.Evaluator"],
        trial: Optional["Trial"],
        max_epoch: int = 128,
        validate_epoch: int = 5,
        score_degradation_max: int = 5,
    ) -> None:
        from optuna.exceptions import TrialPruned

        self.start_learning()
        best_score = -float("inf")
        n_score_degradation = 0
        pb = progress_bar(range(max_epoch))
        for epoch in pb:
            self.run_epoch()
            if (epoch + 1) % validate_epoch:
                continue

            if evaluator is None:
                continue

            target_score = evaluator.get_target_score(self)

            pb.comment = f"{evaluator.target_metric_name}={target_score}"
            relevant_score = target_score
            if relevant_score > best_score:
                best_score = relevant_score
                self.save_state()
                self.learnt_config["train_epochs"] = epoch + 1
                n_score_degradation = 0
            else:
                n_score_degradation += 1
                if n_score_degradation >= score_degradation_max:
                    pb.on_interrupt()
                    break
            if trial is not None:
                trial.report(-relevant_score, epoch)
                if trial.should_prune():
                    pb.on_interrupt()
                    raise TrialPruned()

        if evaluator is not None:
            self.load_state()



import os
from pathlib import Path
from typing import Any, List, Tuple

from pybind11.setup_helpers import Pybind11Extension, build_ext
from setuptools import find_packages, setup

SETUP_DIRECTORY = Path(__file__).resolve().parent

with (SETUP_DIRECTORY / "Readme.md").open() as ifs:
    LONG_DESCRIPTION = ifs.read()

install_requires = (
    [
        "numpy>=1.12.0",
        "fastprogress>=0.2",
        "optuna>=2.5.0",
        "pandas>=1.0.0",
        "scipy>=1.0",
        "colorlog>=4",
        "pydantic>=1.8.2",
        "typing_extensions>=3.10",
    ],
)


class get_eigen_include(object):
    EIGEN3_URL = "https://gitlab.com/libeigen/eigen/-/archive/3.3.7/eigen-3.3.7.zip"
    EIGEN3_DIRNAME = "eigen-3.3.7"

    def __str__(self) -> str:
        eigen_include_dir = os.environ.get("EIGEN3_INCLUDE_DIR", None)

        if eigen_include_dir is not None:
            return eigen_include_dir

        target_dir = SETUP_DIRECTORY / self.EIGEN3_DIRNAME
        if target_dir.exists():
            return target_dir.name

        download_target_dir = SETUP_DIRECTORY / "eigen3.zip"
        import zipfile

        import requests

        response = requests.get(self.EIGEN3_URL, stream=True)
        with download_target_dir.open("wb") as ofs:
            for chunk in response.iter_content(chunk_size=1024):
                ofs.write(chunk)

        with zipfile.ZipFile(download_target_dir) as ifs:
            ifs.extractall()

        return target_dir.name


module_name_and_sources: List[Tuple[str, List[str]]] = [
    ("irspack.evaluation._core", ["cpp_source/evaluator.cpp"]),
    ("irspack.recommenders._ials", ["cpp_source/als/wrapper.cpp"]),
    ("irspack.recommenders._knn", ["cpp_source/knn/wrapper.cpp"]),
    ("irspack.utils._util_cpp", ["cpp_source/util.cpp"]),
]
ext_modules = [
    Pybind11Extension(module_name, sources, include_dirs=[get_eigen_include()])
    for module_name, sources in module_name_and_sources
]


def local_scheme(version: Any) -> str:
    return ""


setup(
    name="irspack",
    # version=get_version(),
    url="https://irspack.readthedocs.io/",
    use_scm_version={
        "local_scheme": local_scheme
    },  # https://github.com/pypa/setuptools_scm/issues/342
    author="Tomoki Ohtsuki",
    author_email="tomoki.otsuki129@gmail.com",
    description="Implicit feedback-based recommender systems, packed for practitioners.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    ext_modules=ext_modules,
    install_requires=install_requires,
    include_package_data=True,
    cmdclass={"build_ext": build_ext},
    packages=find_packages("src"),
    python_requires=">=3.7",
    package_dir={"": "src"},
)



import optuna

from irspack import Evaluator, rowwise_train_test_split
from irspack.recommenders.ials import IALSRecommender
from irspack.utils.sample_data import mf_example_data


def test_doubling_dimension_strategy() -> None:
    X = mf_example_data(100, 100, 8, random_state=0)
    storage = optuna.storages.RDBStorage("sqlite:///:memory:")
    X_train, X_val = rowwise_train_test_split(X, random_state=0)
    SCALE = 1.1
    bp, df = IALSRecommender.tune_doubling_dimension(
        X_train,
        Evaluator(X_val),
        2,
        8,
        n_trials_initial=10,
        n_startup_trials_following=5,
        neighborhood_scale=SCALE,
        storage=storage,
        max_epoch=2,
        random_seed=0,
    )
    tried_dimensions = sorted(df["n_components"].unique())
    assert len(tried_dimensions) == 3
    assert tried_dimensions[0] == 2
    assert tried_dimensions[1] == 4
    assert tried_dimensions[2] == 8

    study_name_2 = storage.get_study_name_from_id(1)
    assert study_name_2.endswith("_2")
    bp_2 = optuna.load_study(study_name_2, storage).best_params

    trials_dim_4 = df[df["n_components"] == 4]
    for dim_4_single_trials in trials_dim_4.itertuples():
        alpha0 = dim_4_single_trials.alpha0
        assert alpha0 <= (bp_2["alpha0"] * SCALE)
        assert alpha0 >= (bp_2["alpha0"] / SCALE)

        reg = dim_4_single_trials.reg
        assert reg <= (bp_2["reg"] * SCALE)
        assert reg >= (bp_2["reg"] / SCALE)

    study_name_4 = storage.get_study_name_from_id(2)
    assert study_name_4.endswith("_4")
    bp_4 = optuna.load_study(study_name_4, storage).best_params

    trials_dim_8 = df[df["n_components"] == 8]
    for dim_8_single_trials in trials_dim_8.itertuples():
        alpha0 = dim_8_single_trials.alpha0
        assert alpha0 <= (bp_4["alpha0"] * SCALE)
        assert alpha0 >= (bp_4["alpha0"] / SCALE)

        reg = dim_8_single_trials.reg
        assert reg <= (bp_4["reg"] * SCALE)
        assert reg >= (bp_4["reg"] / SCALE)

    assert bp["n_components"] == df.sort_values("value").iloc[0]["n_components"]

    assert bp["train_epochs"] <= 2



from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution("irspack").version
except DistributionNotFound:  # pragma: no cover
    # package is not installed
    pass  # pragma: no cover

from .definitions import DenseScoreArray, InteractionMatrix, UserIndexArray
from .evaluation import *
from .recommenders import *
from .split import *
from .utils import *
